from itertools import pairwise, product

import numpy as np
import pandas as pd
import pymc as pm
import pytensor
import pytensor.tensor as pt
import pytest
import statsmodels.api as sm

from numpy.testing import assert_allclose, assert_array_less
from pymc.model.transform.optimization import freeze_dims_and_data

from pymc_extras.statespace import BayesianVARMAX
from pymc_extras.statespace.utils.constants import SHORT_NAME_TO_LONG
from tests.statespace.shared_fixtures import (  # pylint: disable=unused-import
    rng,
)

floatX = pytensor.config.floatX
ps = [0, 1, 2, 3]
qs = [0, 1, 2, 3]
orders = list(product(ps, qs))[1:]
ids = [f"p={x[0]}, q={x[1]}" for x in orders]


@pytest.fixture(scope="session")
def data():
    df = pd.read_csv(
        "tests/statespace/_data/statsmodels_macrodata_processed.csv",
        index_col=0,
        parse_dates=True,
    ).astype(floatX)
    df.index.freq = df.index.inferred_freq
    return df


@pytest.fixture(scope="session")
def varma_mod(data):
    return BayesianVARMAX(
        endog_names=data.columns,
        order=(2, 0),
        stationary_initialization=True,
        verbose=False,
        measurement_error=True,
    )


@pytest.fixture(scope="session")
def pymc_mod(varma_mod, data):
    with pm.Model(coords=varma_mod.coords) as pymc_mod:
        # x0 = pm.Normal("x0", dims=["state"])
        # P0_diag = pm.Exponential("P0_diag", 1, size=varma_mod.k_states)
        # P0 = pm.Deterministic(
        #     "P0", pt.diag(P0_diag), dims=["state", "state_aux"]
        # )
        state_chol, *_ = pm.LKJCholeskyCov(
            "state_chol", n=varma_mod.k_posdef, eta=1, sd_dist=pm.Exponential.dist(1)
        )
        ar_params = pm.Normal(
            "ar_params", mu=0, sigma=0.1, dims=["observed_state", "lag_ar", "observed_state_aux"]
        )
        state_cov = pm.Deterministic(
            "state_cov", state_chol @ state_chol.T, dims=["shock", "shock_aux"]
        )
        sigma_obs = pm.Exponential("sigma_obs", 1, dims=["observed_state"])

        varma_mod.build_statespace_graph(data=data, save_kalman_filter_outputs_in_idata=True)

    return pymc_mod


@pytest.fixture(scope="session")
def idata(pymc_mod, rng):
    with pymc_mod:
        idata = pm.sample_prior_predictive(draws=10, random_seed=rng)

    return idata


def test_mode_argument():
    # Mode argument should be passed to the parent class
    mod = BayesianVARMAX(k_endog=2, order=(3, 0), mode="FAST_RUN", verbose=False)
    assert mod.mode == "FAST_RUN"


@pytest.mark.parametrize("order", orders, ids=ids)
@pytest.mark.parametrize("var", ["AR", "MA", "state_cov"])
@pytest.mark.filterwarnings("ignore::statsmodels.tools.sm_exceptions.EstimationWarning")
def test_VARMAX_param_counts_match_statsmodels(data, order, var):
    p, q = order

    mod = BayesianVARMAX(k_endog=data.shape[1], order=(p, q), verbose=False)
    sm_var = sm.tsa.VARMAX(data, order=(p, q))

    count = mod.param_counts[var]
    if var == "state_cov":
        # Statsmodels only counts the lower triangle
        count = mod.k_posdef * (mod.k_posdef - 1)
    assert count == sm_var.parameters[var.lower()]


@pytest.mark.parametrize("order", orders, ids=ids)
@pytest.mark.filterwarnings("ignore::statsmodels.tools.sm_exceptions.EstimationWarning")
@pytest.mark.filterwarnings("ignore::FutureWarning")
def test_VARMAX_update_matches_statsmodels(data, order, rng):
    p, q = order

    sm_var = sm.tsa.VARMAX(data, order=(p, q))

    param_counts = [None, *np.cumsum(list(sm_var.parameters.values())).tolist()]
    param_slices = [slice(a, b) for a, b in pairwise(param_counts)]
    param_lists = [trend, ar, ma, reg, state_cov, obs_cov] = [
        sm_var.param_names[idx] for idx in param_slices
    ]
    param_d = {
        k: getattr(np, floatX)(rng.normal(scale=0.1) ** 2)
        for param_list in param_lists
        for k in param_list
    }

    res = sm_var.fit_constrained(param_d)

    mod = BayesianVARMAX(
        k_endog=data.shape[1],
        order=(p, q),
        verbose=False,
        measurement_error=False,
        stationary_initialization=False,
    )

    ar_shape = (mod.k_endog, mod.p, mod.k_endog)
    ma_shape = (mod.k_endog, mod.q, mod.k_endog)

    with pm.Model() as pm_mod:
        x0 = pm.Deterministic("x0", pt.zeros(mod.k_states, dtype=floatX))
        P0 = pm.Deterministic("P0", pt.eye(mod.k_states, dtype=floatX))
        ma_params = pm.Deterministic(
            "ma_params",
            pt.as_tensor_variable(np.array([param_d[var] for var in ma])).reshape(ma_shape),
        )
        ar_params = pm.Deterministic(
            "ar_params",
            pt.as_tensor_variable(np.array([param_d[var] for var in ar])).reshape(ar_shape),
        )
        state_chol = np.zeros((mod.k_posdef, mod.k_posdef), dtype=floatX)
        state_chol[np.tril_indices(mod.k_posdef)] = np.array([param_d[var] for var in state_cov])
        state_cov = pm.Deterministic("state_cov", pt.as_tensor_variable(state_chol @ state_chol.T))
        mod._insert_random_variables()

        matrices = pm.draw(mod.subbed_ssm)
        matrix_dict = dict(zip(SHORT_NAME_TO_LONG.values(), matrices))

    for matrix in ["transition", "selection", "state_cov", "obs_cov", "design"]:
        assert_allclose(matrix_dict[matrix], sm_var.ssm[matrix])


@pytest.mark.parametrize("filter_output", ["filtered", "predicted", "smoothed"])
def test_all_prior_covariances_are_PSD(filter_output, pymc_mod, rng):
    rv = pymc_mod[f"{filter_output}_covariances"]
    cov_mats = pm.draw(rv, 100, random_seed=rng)
    w, v = np.linalg.eig(cov_mats)
    assert_array_less(0, w, err_msg=f"Smallest eigenvalue: {min(w.ravel())}")


parameters = [
    {"n_steps": 10, "shock_size": None},
    {"n_steps": 10, "shock_size": 1.0},
    {"n_steps": 10, "shock_size": np.array([1.0, 0.0, 0.0])},
    {
        "n_steps": 10,
        "shock_cov": np.array([[1.38, 0.58, -1.84], [0.58, 0.99, -0.82], [-1.84, -0.82, 2.51]]),
    },
    {
        "shock_trajectory": np.r_[
            np.zeros((3, 3), dtype=floatX),
            np.array([[1.0, 0.0, 0.0]]).astype(floatX),
            np.zeros((6, 3), dtype=floatX),
        ]
    },
]

ids = ["from-posterior-cov", "scalar_shock_size", "array_shock_size", "user-cov", "trajectory"]


@pytest.mark.parametrize("parameters", parameters, ids=ids)
@pytest.mark.skipif(floatX == "float32", reason="Impulse covariance not PSD if float32")
def test_impulse_response(parameters, varma_mod, idata, rng):
    irf = varma_mod.impulse_response_function(idata.prior, random_seed=rng, **parameters)

    assert not np.any(np.isnan(irf.irf.values))


class TestVARMAXWithExogenous:
    def test_create_varmax_with_exogenous_k_exog_int(self, data):
        mod = BayesianVARMAX(
            k_endog=data.shape[1],
            order=(1, 0),
            k_exog=2,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == 2
        assert mod.exog_state_names == ["exogenous_0", "exogenous_1"]
        assert mod.data_names == ["exogenous_data"]
        assert mod.param_dims["beta_exog"] == ("observed_state", "exogenous")
        assert mod.coords["exogenous"] == ["exogenous_0", "exogenous_1"]
        assert mod.param_info["beta_exog"]["shape"] == (mod.k_endog, 2)
        assert mod.param_info["beta_exog"]["dims"] == ("observed_state", "exogenous")

    def test_create_varmax_with_exogenous_list_of_names(self, data):
        mod = BayesianVARMAX(
            k_endog=data.shape[1],
            order=(1, 0),
            exog_state_names=["foo", "bar"],
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == 2
        assert mod.exog_state_names == ["foo", "bar"]
        assert mod.data_names == ["exogenous_data"]
        assert mod.param_dims["beta_exog"] == ("observed_state", "exogenous")
        assert mod.coords["exogenous"] == ["foo", "bar"]
        assert mod.param_info["beta_exog"]["shape"] == (mod.k_endog, 2)
        assert mod.param_info["beta_exog"]["dims"] == ("observed_state", "exogenous")

    def test_create_varmax_with_exogenous_both_defined_correctly(self, data):
        mod = BayesianVARMAX(
            k_endog=data.shape[1],
            order=(1, 0),
            k_exog=2,
            exog_state_names=["a", "b"],
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == 2
        assert mod.exog_state_names == ["a", "b"]
        assert mod.data_names == ["exogenous_data"]
        assert mod.param_dims["beta_exog"] == ("observed_state", "exogenous")
        assert mod.coords["exogenous"] == ["a", "b"]
        assert mod.param_info["beta_exog"]["shape"] == (mod.k_endog, 2)
        assert mod.param_info["beta_exog"]["dims"] == ("observed_state", "exogenous")

    def test_create_varmax_with_exogenous_k_exog_dict(self, data):
        k_exog = {"observed_0": 2, "observed_1": 1, "observed_2": 0}
        mod = BayesianVARMAX(
            endog_names=["observed_0", "observed_1", "observed_2"],
            order=(1, 0),
            k_exog=k_exog,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == k_exog
        assert mod.exog_state_names == {
            "observed_0": ["observed_0_exogenous_0", "observed_0_exogenous_1"],
            "observed_1": ["observed_1_exogenous_0"],
            "observed_2": [],
        }
        assert mod.data_names == [
            "observed_0_exogenous_data",
            "observed_1_exogenous_data",
            "observed_2_exogenous_data",
        ]
        assert mod.param_dims["beta_observed_0"] == ("exogenous_observed_0",)
        assert mod.param_dims["beta_observed_1"] == ("exogenous_observed_1",)
        assert (
            "beta_observed_2" not in mod.param_dims
            or mod.param_info.get("beta_observed_2") is None
            or mod.param_info.get("beta_observed_2", {}).get("shape", (0,))[0] == 0
        )

        assert mod.coords["exogenous_observed_0"] == [
            "observed_0_exogenous_0",
            "observed_0_exogenous_1",
        ]
        assert mod.coords["exogenous_observed_1"] == ["observed_1_exogenous_0"]
        assert "exogenous_observed_2" in mod.coords and mod.coords["exogenous_observed_2"] == []

        assert mod.param_info["beta_observed_0"]["shape"] == (2,)
        assert mod.param_info["beta_observed_0"]["dims"] == ("exogenous_observed_0",)
        assert mod.param_info["beta_observed_1"]["shape"] == (1,)
        assert mod.param_info["beta_observed_1"]["dims"] == ("exogenous_observed_1",)

    def test_create_varmax_with_exogenous_exog_names_dict(self, data):
        exog_state_names = {"observed_0": ["a", "b"], "observed_1": ["c"], "observed_2": []}
        mod = BayesianVARMAX(
            endog_names=["observed_0", "observed_1", "observed_2"],
            order=(1, 0),
            exog_state_names=exog_state_names,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == {"observed_0": 2, "observed_1": 1, "observed_2": 0}
        assert mod.exog_state_names == exog_state_names
        assert mod.data_names == [
            "observed_0_exogenous_data",
            "observed_1_exogenous_data",
            "observed_2_exogenous_data",
        ]
        assert mod.param_dims["beta_observed_0"] == ("exogenous_observed_0",)
        assert mod.param_dims["beta_observed_1"] == ("exogenous_observed_1",)
        assert (
            "beta_observed_2" not in mod.param_dims
            or mod.param_info.get("beta_observed_2") is None
            or mod.param_info.get("beta_observed_2", {}).get("shape", (0,))[0] == 0
        )

        assert mod.coords["exogenous_observed_0"] == ["a", "b"]
        assert mod.coords["exogenous_observed_1"] == ["c"]
        assert "exogenous_observed_2" in mod.coords and mod.coords["exogenous_observed_2"] == []

        assert mod.param_info["beta_observed_0"]["shape"] == (2,)
        assert mod.param_info["beta_observed_0"]["dims"] == ("exogenous_observed_0",)
        assert mod.param_info["beta_observed_1"]["shape"] == (1,)
        assert mod.param_info["beta_observed_1"]["dims"] == ("exogenous_observed_1",)

    def test_create_varmax_with_exogenous_both_dict_correct(self, data):
        k_exog = {"observed_0": 2, "observed_1": 1}
        exog_state_names = {"observed_0": ["a", "b"], "observed_1": ["c"]}
        mod = BayesianVARMAX(
            endog_names=["observed_0", "observed_1"],
            order=(1, 0),
            k_exog=k_exog,
            exog_state_names=exog_state_names,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == k_exog
        assert mod.exog_state_names == exog_state_names
        assert mod.data_names == ["observed_0_exogenous_data", "observed_1_exogenous_data"]
        assert mod.param_dims["beta_observed_0"] == ("exogenous_observed_0",)
        assert mod.param_dims["beta_observed_1"] == ("exogenous_observed_1",)
        assert mod.coords["exogenous_observed_0"] == ["a", "b"]
        assert mod.coords["exogenous_observed_1"] == ["c"]
        assert mod.param_info["beta_observed_0"]["shape"] == (2,)
        assert mod.param_info["beta_observed_0"]["dims"] == ("exogenous_observed_0",)
        assert mod.param_info["beta_observed_1"]["shape"] == (1,)
        assert mod.param_info["beta_observed_1"]["dims"] == ("exogenous_observed_1",)

    def test_create_varmax_with_exogenous_dict_converts_to_list(self, data):
        exog_state_names = {
            "observed_0": ["a", "b"],
            "observed_1": ["a", "b"],
            "observed_2": ["a", "b"],
        }
        mod = BayesianVARMAX(
            endog_names=["observed_0", "observed_1", "observed_2"],
            order=(1, 0),
            exog_state_names=exog_state_names,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )

        assert mod.k_exog == 2
        assert mod.exog_state_names == ["a", "b"]
        assert mod.data_names == ["exogenous_data"]
        assert mod.param_dims["beta_exog"] == ("observed_state", "exogenous")
        assert mod.coords["exogenous"] == ["a", "b"]
        assert mod.param_info["beta_exog"]["shape"] == (mod.k_endog, 2)
        assert mod.param_info["beta_exog"]["dims"] == ("observed_state", "exogenous")

    def test_create_varmax_with_exogenous_raises_if_args_disagree(self, data):
        # List case
        with pytest.raises(
            ValueError, match="Length of exog_state_names does not match provided k_exog"
        ):
            BayesianVARMAX(
                k_endog=2,
                order=(1, 0),
                k_exog=3,
                exog_state_names=["a", "b"],
                verbose=False,
                measurement_error=False,
                stationary_initialization=False,
            )

        # Dict case
        with pytest.raises(
            ValueError,
            match="If k_exog is an int, exog_state_names must be a list of the same length",
        ):
            BayesianVARMAX(
                k_endog=2,
                order=(1, 0),
                k_exog=2,
                exog_state_names={"observed_0": ["a"], "observed_1": ["b"]},
                verbose=False,
                measurement_error=False,
                stationary_initialization=False,
            )

        # dict + list
        with pytest.raises(
            ValueError, match="If k_exog is a dict, exog_state_names must be a dict as well"
        ):
            BayesianVARMAX(
                endog_names=["observed_0", "observed_1"],
                order=(1, 0),
                k_exog={"observed_0": 1, "observed_1": 1},
                exog_state_names=["a", "b"],
                verbose=False,
                measurement_error=False,
                stationary_initialization=False,
            )

        # Dict/dict, key mismatch
        with pytest.raises(
            ValueError, match="Keys of k_exog and exog_state_names dicts must match"
        ):
            BayesianVARMAX(
                endog_names=["observed_0", "observed_1"],
                order=(1, 0),
                k_exog={"observed_0": 1, "observed_1": 1},
                exog_state_names={"observed_0": ["a"], "observed_2": ["b"]},
                verbose=False,
                measurement_error=False,
                stationary_initialization=False,
            )

        # Dict/dict, length mismatch
        with pytest.raises(ValueError, match="lengths of exog_state_names lists must match"):
            BayesianVARMAX(
                endog_names=["observed_0", "observed_1"],
                order=(1, 0),
                k_exog={"observed_0": 2, "observed_1": 1},
                exog_state_names={"observed_0": ["a"], "observed_1": ["b"]},
                verbose=False,
                measurement_error=False,
                stationary_initialization=False,
            )

    @pytest.mark.parametrize(
        "k_exog, exog_state_names",
        [
            (2, None),
            (None, ["foo", "bar"]),
            (None, {"y1": ["a", "b"], "y2": ["c"]}),
        ],
        ids=["k_exog_int", "exog_state_names_list", "exog_state_names_dict"],
    )
    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_varmax_with_exog(self, rng, k_exog, exog_state_names):
        endog_names = ["y1", "y2", "y3"]
        n_obs = 50
        time_idx = pd.date_range(start="2020-01-01", periods=n_obs, freq="D")

        y = rng.normal(size=(n_obs, len(endog_names)))
        df = pd.DataFrame(y, columns=endog_names, index=time_idx).astype(floatX)

        if isinstance(exog_state_names, dict):
            exog_data = {
                f"{name}_exogenous_data": pd.DataFrame(
                    rng.normal(size=(n_obs, len(exog_names))).astype(floatX),
                    columns=exog_names,
                    index=time_idx,
                )
                for name, exog_names in exog_state_names.items()
            }
        else:
            exog_names = exog_state_names or [f"exogenous_{i}" for i in range(k_exog)]
            exog_data = {
                "exogenous_data": pd.DataFrame(
                    rng.normal(size=(n_obs, k_exog or len(exog_state_names))).astype(floatX),
                    columns=exog_names,
                    index=time_idx,
                )
            }

        mod = BayesianVARMAX(
            endog_names=endog_names,
            order=(1, 0),
            k_exog=k_exog,
            exog_state_names=exog_state_names,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
            mode="JAX",
        )

        with pm.Model(coords=mod.coords) as m:
            for var_name, data in exog_data.items():
                pm.Data(var_name, data, dims=mod.data_info[var_name]["dims"])

            x0 = pm.Deterministic("x0", pt.zeros(mod.k_states), dims=mod.param_dims["x0"])
            P0_diag = pm.Exponential("P0_diag", 1.0, dims=mod.param_dims["P0"][0])
            P0 = pm.Deterministic("P0", pt.diag(P0_diag), dims=mod.param_dims["P0"])

            ar_params = pm.Normal("ar_params", mu=0, sigma=1, dims=mod.param_dims["ar_params"])
            state_cov_diag = pm.Exponential(
                "state_cov_diag", 1.0, dims=mod.param_dims["state_cov"][0]
            )
            state_cov = pm.Deterministic(
                "state_cov", pt.diag(state_cov_diag), dims=mod.param_dims["state_cov"]
            )

            # Exogenous priors
            if isinstance(mod.exog_state_names, list):
                beta_exog = pm.Normal("beta_exog", mu=0, sigma=1, dims=mod.param_dims["beta_exog"])
            elif isinstance(mod.exog_state_names, dict):
                for name in mod.exog_state_names:
                    if mod.exog_state_names.get(name):
                        pm.Normal(
                            f"beta_{name}", mu=0, sigma=1, dims=mod.param_dims[f"beta_{name}"]
                        )

            mod.build_statespace_graph(data=df)

        with freeze_dims_and_data(m):
            prior = pm.sample_prior_predictive(
                draws=10, random_seed=rng, compile_kwargs={"mode": "JAX"}
            )

        prior_cond = mod.sample_conditional_prior(prior, mvn_method="eigh")
        beta_dot_data = prior_cond.filtered_prior_observed.values - prior_cond.filtered_prior.values

        if isinstance(exog_state_names, list) or k_exog is not None:
            beta = prior.prior.beta_exog
            assert beta.shape == (1, 10, 3, 2)

            np.testing.assert_allclose(
                beta_dot_data,
                np.einsum("tx,...sx->...ts", exog_data["exogenous_data"].values, beta),
                atol=1e-2,
            )

        elif isinstance(exog_state_names, dict):
            assert prior.prior.beta_y1.shape == (1, 10, 2)
            assert prior.prior.beta_y2.shape == (1, 10, 1)

            obs_intercept = [
                np.einsum("tx,...x->...t", exog_data[f"{name}_exogenous_data"].values, beta)
                for name, beta in zip(["y1", "y2"], [prior.prior.beta_y1, prior.prior.beta_y2])
            ]

            # y3 has no exogenous variables
            obs_intercept.append(np.zeros_like(obs_intercept[0]))

            np.testing.assert_allclose(beta_dot_data, np.stack(obs_intercept, axis=-1), atol=1e-2)



import numpy as np
import pytest

from pytensor import function
from pytensor import tensor as pt

from pymc_extras.statespace.models.utilities import (
    add_tensors_by_dim_labels,
    join_tensors_by_dim_labels,
    reorder_from_labels,
)


def test_reorder_from_labels():
    x = pt.tensor("x", shape=(None, None))
    labels = ["A", "B", "D"]
    combined_labels = ["A", "D", "B"]

    x_sorted = reorder_from_labels(x, labels, combined_labels, labeled_axis=0)
    fn = function([x], x_sorted)

    test_val = np.eye(3) * np.arange(1, 4)
    idx = np.array([0, 2, 1])
    out = fn(test_val)
    np.testing.assert_allclose(out, test_val[idx, :])

    x_sorted = reorder_from_labels(x, labels, combined_labels, labeled_axis=1)
    fn = function([x], x_sorted)

    out = fn(test_val)
    np.testing.assert_allclose(out, test_val[:, idx])

    x_sorted = reorder_from_labels(x, labels, combined_labels, labeled_axis=(0, 1))
    fn = function([x], x_sorted)

    out = fn(test_val)
    np.testing.assert_allclose(out, test_val[np.ix_(idx, idx)])


def make_zeros(x):
    if x.ndim == 1:
        zeros = np.zeros(
            1,
        )
    else:
        zeros = np.zeros((x.shape[0], 1))
    return zeros


def add(left, right):
    return left + right


def same_but_mixed(left, right):
    return left + right[..., np.array([1, 2, 0])]


def concat(left, right):
    return np.concatenate([left, right], axis=-1)


def pad_and_add_left(left, right):
    left = np.concatenate([left, make_zeros(left)], axis=-1)
    return left + right


def pad_and_add_right(left, right):
    right = np.concatenate([right, make_zeros(right)], axis=-1)
    return left + right


def mixed_and_padded(left, right):
    left = np.concatenate([left, make_zeros(left)], axis=-1)
    right = right[..., np.array([2, 1, 0])]
    return left + right


@pytest.mark.parametrize(
    "left_names, right_names, expected_computation",
    [
        (["data"], ["data"], add),
        (["A", "C", "B"], ["B", "A", "C"], same_but_mixed),
        (["data"], ["different_data"], concat),
        (["data"], ["data", "different_data"], pad_and_add_left),
        (["data", "more_data"], ["data"], pad_and_add_right),
        (["A", "B"], ["D", "B", "A"], mixed_and_padded),
    ],
    ids=[
        "same_names",
        "same_but_mixed",
        "different_names",
        "overlap_right",
        "overlap_left",
        "pad_and_mix",
    ],
)
@pytest.mark.parametrize("ndim", [1, 2], ids=["vector", "matrix"])
def test_add_matrices_by_observed_state_names(left_names, right_names, expected_computation, ndim):
    rng = np.random.default_rng()
    n_left = len(left_names)
    n_right = len(right_names)

    left = pt.tensor("left", shape=(None,) * ndim)
    right = pt.tensor("right", shape=(None,) * ndim)

    result = add_tensors_by_dim_labels(left, right, left_names, right_names)
    fn = function([left, right], result)

    left_value = rng.normal(size=(n_left,) if ndim == 1 else (10, n_left))
    right_value = rng.normal(size=(n_right,) if ndim == 1 else (10, n_right))

    np.testing.assert_allclose(
        fn(left_value, right_value), expected_computation(left_value, right_value)
    )


class TestAddCovarianceMatrices:
    def _setup_H(self, states_1, states_2):
        n_1 = len(states_1)
        n_2 = len(states_2)

        H_1 = pt.tensor("H_1", shape=(n_1, n_1))
        H_2 = pt.tensor("H_2", shape=(n_2, n_2))

        return H_1, H_2

    @pytest.mark.parametrize("n_states", [1, 3], ids=["1x1", "3x3"])
    def test_add_fully_overlapping_covariance_matrices(self, n_states):
        rng = np.random.default_rng()
        states = list("ABCD")

        observed_states_1 = states[:n_states]
        observed_states_2 = states[:n_states]

        H_1, H_2 = self._setup_H(observed_states_1, observed_states_2)
        res = add_tensors_by_dim_labels(
            H_1, H_2, observed_states_1, observed_states_2, labeled_axis=(0, 1)
        )

        fn = function([H_1, H_2], res)

        H_1_val = rng.normal(size=(n_states, n_states))
        H_2_val = rng.normal(size=(n_states, n_states))

        np.testing.assert_allclose(fn(H_1_val, H_2_val), H_1_val + H_2_val)

    def test_add_fully_overlapping_mixed_covariance_matrices(self):
        rng = np.random.default_rng()

        observed_states_1 = ["A", "B", "C", "D"]
        observed_states_2 = ["A", "B", "C", "D"]
        rng.shuffle(observed_states_2)

        H_1, H_2 = self._setup_H(observed_states_1, observed_states_2)

        res = add_tensors_by_dim_labels(
            H_1, H_2, observed_states_1, observed_states_2, labeled_axis=(0, 1)
        )

        H_1_val = rng.normal(size=(4, 4))
        H_2_val = rng.normal(size=(4, 4))

        fn = function([H_1, H_2], res)

        state_to_idx = {name: idx for idx, name in enumerate(observed_states_1)}
        idx = np.argsort([state_to_idx[state] for state in observed_states_2])

        np.testing.assert_allclose(fn(H_1_val, H_2_val), H_1_val + H_2_val[np.ix_(idx, idx)])

    def test_add_non_overlapping_covaraince_matrices(self):
        rng = np.random.default_rng()

        observed_states_1 = ["A", "B"]
        observed_states_2 = ["C", "D"]

        H_1, H_2 = self._setup_H(observed_states_1, observed_states_2)

        res = add_tensors_by_dim_labels(
            H_1, H_2, observed_states_1, observed_states_2, labeled_axis=(0, 1)
        )

        H_1_val = rng.normal(size=(2, 2))
        H_2_val = rng.normal(size=(2, 2))
        zeros = np.zeros_like(H_1_val)

        fn = function([H_1, H_2], res)

        np.testing.assert_allclose(
            fn(H_1_val, H_2_val), np.block([[H_1_val, zeros], [zeros, H_2_val]])
        )

    def test_add_partially_overlapping_covaraince_matrices(self):
        rng = np.random.default_rng()
        observed_states_1 = ["A", "B"]
        observed_states_2 = ["B", "C", "D", "A"]
        H_1, H_2 = self._setup_H(observed_states_1, observed_states_2)

        res = add_tensors_by_dim_labels(
            H_1, H_2, observed_states_1, observed_states_2, labeled_axis=(-2, -1)
        )

        fn = function([H_1, H_2], res)
        H_1_val = rng.normal(size=(2, 2))
        H_2_val = rng.normal(size=(4, 4))

        upper = np.zeros((4, 4))
        upper_idx = np.ix_([0, 1], [0, 1])
        upper[upper_idx] = H_1_val
        expected_value = upper + H_2_val[np.ix_([3, 0, 1, 2], [3, 0, 1, 2])]

        np.testing.assert_allclose(fn(H_1_val, H_2_val), expected_value)


class TestJoinDesignMatrices:
    def _setup_Z(self, states_1, states_2, k_endog=2):
        Z_1 = pt.tensor("Z_1", shape=(len(states_1), k_endog))
        Z_2 = pt.tensor("Z_2", shape=(len(states_2), k_endog))

        return Z_1, Z_2

    def test_join_fully_overlapping_design_matrices(self):
        observed_states_1 = ["A"]
        observed_states_2 = ["A"]

        Z_1, Z_2 = self._setup_Z(observed_states_1, observed_states_2)
        res = join_tensors_by_dim_labels(
            Z_1, Z_2, observed_states_1, observed_states_2, labeled_axis=0, join_axis=1
        )

        fn = function([Z_1, Z_2], res)

        Z_1_val = np.array([[1.0, 0.0]])
        Z_2_val = np.array([[0.0, 1.0]])

        np.testing.assert_allclose(fn(Z_1_val, Z_2_val), np.array([[1.0, 0.0, 0.0, 1.0]]))

    def test_join_fully_overlapping_mixed_design_matrices(self):
        observed_states_1 = ["A", "B", "C"]
        observed_states_2 = ["C", "B", "A"]

        Z_1, Z_2 = self._setup_Z(observed_states_1, observed_states_2, k_endog=3)
        res = join_tensors_by_dim_labels(
            Z_1, Z_2, observed_states_1, observed_states_2, labeled_axis=0, join_axis=1
        )

        fn = function([Z_1, Z_2], res)

        Z_1_val = np.array([[1.0, 0.0, 1.0], [0.0, 1.0, 0.0], [1.0, 0.0, 1.0]])
        Z_2_val = np.array([[0.0, 1.0, 1.0], [1.0, 0.0, 1.0], [1.0, 0.0, 0.0]])

        # Rows 0 and 2 should be swapped in the output, because the ordering A, B, C becomes canonical as it was passed
        # in first, and because we said the labeled dim was axis=0. After reordering, the matrices should be
        # concatenated on axis = 1 (again, as requested).
        np.testing.assert_allclose(
            fn(Z_1_val, Z_2_val),
            np.array(
                [
                    [1.0, 0.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
                    [1.0, 0.0, 1.0, 0.0, 1.0, 1.0],
                ]
            ),
        )

    def test_join_non_overlapping_design_matrices(self):
        observed_states_1 = ["A"]
        observed_states_2 = ["B"]

        Z_1, Z_2 = self._setup_Z(observed_states_1, observed_states_2)
        fn = function(
            [Z_1, Z_2], join_tensors_by_dim_labels(Z_1, Z_2, observed_states_1, observed_states_2)
        )

        Z_1_val = np.array([[1.0, 0.0]])
        Z_2_val = np.array([[1.0, 0.0]])
        out = fn(Z_1_val, Z_2_val)

        np.testing.assert_allclose(out, [[1.0, 0.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0]])

    def test_join_partially_overlapping_design_matrices(self):
        observed_states_1 = ["A"]
        observed_states_2 = ["A", "B", "C"]

        Z_1, Z_2 = self._setup_Z(observed_states_1, observed_states_2)
        res = join_tensors_by_dim_labels(
            Z_1, Z_2, observed_states_1, observed_states_2, labeled_axis=0, join_axis=1
        )
        fn = function([Z_1, Z_2], res)

        Z_1_val = np.array([[1.0, 0.0]])
        Z_2_val = np.array([[0.0, 1.0], [1.0, 0.0], [1.0, 0.0]])

        # Z_1 should be zero padded with the missing observed states, then concatenated along axis = -1
        expected_output = np.array(
            [[1.0, 0.0, 0.0, 1.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 1.0, 0.0]]
        )

        np.testing.assert_allclose(fn(Z_1_val, Z_2_val), expected_output)



"""Class that represents a prior distribution.

The `Prior` class is a wrapper around PyMC distributions that allows the user
to create outside of the PyMC model.

Examples
--------
Create a normal prior.

.. code-block:: python

    from pymc_extras.prior import Prior

    normal = Prior("Normal")

Create a hierarchical normal prior by using distributions for the parameters
and specifying the dims.

.. code-block:: python

    hierarchical_normal = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="channel",
    )

Create a non-centered hierarchical normal prior with the `centered` parameter.

.. code-block:: python

    non_centered_hierarchical_normal = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="channel",
        # Only change needed to make it non-centered
        centered=False,
    )

Create a hierarchical beta prior by using Beta distribution, distributions for
the parameters, and specifying the dims.

.. code-block:: python

    hierarchical_beta = Prior(
        "Beta",
        alpha=Prior("HalfNormal"),
        beta=Prior("HalfNormal"),
        dims="channel",
    )

Create a transformed hierarchical normal prior by using the `transform`
parameter. Here the "sigmoid" transformation comes from `pm.math`.

.. code-block:: python

    transformed_hierarchical_normal = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        transform="sigmoid",
        dims="channel",
    )

Create a prior with a custom transform function by registering it with
`register_tensor_transform`.

.. code-block:: python

    from pymc_extras.prior import register_tensor_transform

    def custom_transform(x):
        return x ** 2

    register_tensor_transform("square", custom_transform)

    custom_distribution = Prior("Normal", transform="square")

"""

from __future__ import annotations

import copy

from collections.abc import Callable
from functools import partial
from inspect import signature
from typing import Any, Protocol, runtime_checkable

import numpy as np
import pymc as pm
import pytensor.tensor as pt
import xarray as xr

from pydantic import InstanceOf, validate_call
from pydantic.dataclasses import dataclass
from pymc.distributions.shape_utils import Dims

from pymc_extras.deserialize import deserialize, register_deserialization


class UnsupportedShapeError(Exception):
    """Error for when the shapes from variables are not compatible."""


class UnsupportedDistributionError(Exception):
    """Error for when an unsupported distribution is used."""


class UnsupportedParameterizationError(Exception):
    """The follow parameterization is not supported."""


class MuAlreadyExistsError(Exception):
    """Error for when 'mu' is present in Prior."""

    def __init__(self, distribution: Prior) -> None:
        self.distribution = distribution
        self.message = f"The mu parameter is already defined in {distribution}"
        super().__init__(self.message)


class UnknownTransformError(Exception):
    """Error for when an unknown transform is used."""


def _remove_leading_xs(args: list[str | int]) -> list[str | int]:
    """Remove leading 'x' from the args."""
    while args and args[0] == "x":
        args.pop(0)

    return args


def handle_dims(x: pt.TensorLike, dims: Dims, desired_dims: Dims) -> pt.TensorVariable:
    """Take a tensor of dims `dims` and align it to `desired_dims`.

    Doesn't check for validity of the dims

    Examples
    --------
    1D to 2D with new dim

    .. code-block:: python

        x = np.array([1, 2, 3])
        dims = "channel"

        desired_dims = ("channel", "group")

        handle_dims(x, dims, desired_dims)

    """
    x = pt.as_tensor_variable(x)

    if np.ndim(x) == 0:
        return x

    dims = dims if isinstance(dims, tuple) else (dims,)
    desired_dims = desired_dims if isinstance(desired_dims, tuple) else (desired_dims,)

    if difference := set(dims).difference(desired_dims):
        raise UnsupportedShapeError(
            f"Dims {dims} of data are not a subset of the desired dims {desired_dims}. "
            f"{difference} is missing from the desired dims."
        )

    aligned_dims = np.array(dims)[:, None] == np.array(desired_dims)

    missing_dims = aligned_dims.sum(axis=0) == 0
    new_idx = aligned_dims.argmax(axis=0)

    args = ["x" if missing else idx for (idx, missing) in zip(new_idx, missing_dims, strict=False)]
    args = _remove_leading_xs(args)
    return x.dimshuffle(*args)


DimHandler = Callable[[pt.TensorLike, Dims], pt.TensorLike]


def create_dim_handler(desired_dims: Dims) -> DimHandler:
    """Wrap the `handle_dims` function to act like the previous `create_dim_handler` function."""

    def func(x: pt.TensorLike, dims: Dims) -> pt.TensorVariable:
        return handle_dims(x, dims, desired_dims)

    return func


def _dims_to_str(obj: tuple[str, ...]) -> str:
    if len(obj) == 1:
        return f'"{obj[0]}"'

    return "(" + ", ".join(f'"{i}"' if isinstance(i, str) else str(i) for i in obj) + ")"


def _get_pymc_distribution(name: str) -> type[pm.Distribution]:
    if not hasattr(pm, name):
        raise UnsupportedDistributionError(f"PyMC doesn't have a distribution of name {name!r}")

    return getattr(pm, name)


Transform = Callable[[pt.TensorLike], pt.TensorLike]

CUSTOM_TRANSFORMS: dict[str, Transform] = {}


def register_tensor_transform(name: str, transform: Transform) -> None:
    """Register a tensor transform function to be used in the `Prior` class.

    Parameters
    ----------
    name : str
        The name of the transform.
    func : Callable[[pt.TensorLike], pt.TensorLike]
        The function to apply to the tensor.

    Examples
    --------
    Register a custom transform function.

    .. code-block:: python

        from pymc_extras.prior import (
            Prior,
            register_tensor_transform,
        )

        def custom_transform(x):
            return x ** 2

        register_tensor_transform("square", custom_transform)

        custom_distribution = Prior("Normal", transform="square")

    """
    CUSTOM_TRANSFORMS[name] = transform


def _get_transform(name: str):
    if name in CUSTOM_TRANSFORMS:
        return CUSTOM_TRANSFORMS[name]

    for module in (pt, pm.math):
        if hasattr(module, name):
            break
    else:
        module = None

    if not module:
        msg = (
            f"Neither pytensor.tensor nor pymc.math have the function {name!r}. "
            "If this is a custom function, register it with the "
            "`pymc_extras.prior.register_tensor_transform` function before "
            "previous function call."
        )

        raise UnknownTransformError(msg)

    return getattr(module, name)


def _get_pymc_parameters(distribution: pm.Distribution) -> set[str]:
    return set(signature(distribution.dist).parameters.keys()) - {"kwargs", "args"}


@runtime_checkable
class VariableFactory(Protocol):
    """Protocol for something that works like a Prior class."""

    dims: tuple[str, ...]

    def create_variable(self, name: str) -> pt.TensorVariable:
        """Create a TensorVariable."""


def sample_prior(
    factory: VariableFactory,
    coords=None,
    name: str = "variable",
    wrap: bool = False,
    **sample_prior_predictive_kwargs,
) -> xr.Dataset:
    """Sample the prior for an arbitrary VariableFactory.

    Parameters
    ----------
    factory : VariableFactory
        The factory to sample from.
    coords : dict[str, list[str]], optional
        The coordinates for the variable, by default None.
        Only required if the dims are specified.
    name : str, optional
        The name of the variable, by default "variable".
    wrap : bool, optional
        Whether to wrap the variable in a `pm.Deterministic` node, by default False.
    sample_prior_predictive_kwargs : dict
        Additional arguments to pass to `pm.sample_prior_predictive`.

    Returns
    -------
    xr.Dataset
        The dataset of the prior samples.

    Example
    -------
    Sample from an arbitrary variable factory.

    .. code-block:: python

        import pymc as pm

        import pytensor.tensor as pt

        from pymc_extras.prior import sample_prior

        class CustomVariableDefinition:
            def __init__(self, dims, n: int):
                self.dims = dims
                self.n = n

            def create_variable(self, name: str) -> "TensorVariable":
                x = pm.Normal(f"{name}_x", mu=0, sigma=1, dims=self.dims)
                return pt.sum([x ** n for n in range(1, self.n + 1)], axis=0)

        cubic = CustomVariableDefinition(dims=("channel",), n=3)
        coords = {"channel": ["C1", "C2", "C3"]}
        # Doesn't include the return value
        prior = sample_prior(cubic, coords=coords)

        prior_with = sample_prior(cubic, coords=coords, wrap=True)

    """
    coords = coords or {}

    if isinstance(factory.dims, str):
        dims = (factory.dims,)
    else:
        dims = factory.dims

    if missing_keys := set(dims) - set(coords.keys()):
        raise KeyError(f"Coords are missing the following dims: {missing_keys}")

    with pm.Model(coords=coords) as model:
        if wrap:
            pm.Deterministic(name, factory.create_variable(name), dims=factory.dims)
        else:
            factory.create_variable(name)

    return pm.sample_prior_predictive(
        model=model,
        **sample_prior_predictive_kwargs,
    ).prior


class Prior:
    """A class to represent a prior distribution.

    Make use of the various helper methods to understand the distributions
    better.

    - `preliz` attribute to get the equivalent distribution in `preliz`
    - `sample_prior` method to sample from the prior
    - `to_graph` get a dummy model graph with the distribution
    - `constrain` to shift the distribution to a different range

    Parameters
    ----------
    distribution : str
        The name of PyMC distribution.
    dims : Dims, optional
        The dimensions of the variable, by default None
    centered : bool, optional
        Whether the variable is centered or not, by default True.
        Only allowed for Normal distribution.
    transform : str, optional
        The name of the transform to apply to the variable after it is
        created, by default None or no transform. The transformation must
        be registered with `register_tensor_transform` function or
        be available in either `pytensor.tensor` or `pymc.math`.

    """

    # Taken from https://en.wikipedia.org/wiki/Location%E2%80%93scale_family
    non_centered_distributions: dict[str, dict[str, float]] = {
        "Normal": {"mu": 0, "sigma": 1},
        "StudentT": {"mu": 0, "sigma": 1},
        "ZeroSumNormal": {"sigma": 1},
    }

    pymc_distribution: type[pm.Distribution]
    pytensor_transform: Callable[[pt.TensorLike], pt.TensorLike] | None

    @validate_call
    def __init__(
        self,
        distribution: str,
        *,
        dims: Dims | None = None,
        centered: bool = True,
        transform: str | None = None,
        **parameters,
    ) -> None:
        self.distribution = distribution
        self.parameters = parameters
        self.dims = dims
        self.centered = centered
        self.transform = transform

        self._checks()

    @property
    def distribution(self) -> str:
        """The name of the PyMC distribution."""
        return self._distribution

    @distribution.setter
    def distribution(self, distribution: str) -> None:
        if hasattr(self, "_distribution"):
            raise AttributeError("Can't change the distribution")

        self._distribution = distribution
        self.pymc_distribution = _get_pymc_distribution(distribution)

    @property
    def transform(self) -> str | None:
        """The name of the transform to apply to the variable after it is created."""
        return self._transform

    @transform.setter
    def transform(self, transform: str | None) -> None:
        self._transform = transform
        self.pytensor_transform = not transform or _get_transform(transform)  # type: ignore

    @property
    def dims(self) -> Dims:
        """The dimensions of the variable."""
        return self._dims

    @dims.setter
    def dims(self, dims) -> None:
        if isinstance(dims, str):
            dims = (dims,)

        if isinstance(dims, list):
            dims = tuple(dims)

        self._dims = dims or ()

        self._param_dims_work()
        self._unique_dims()

    def __getitem__(self, key: str) -> Prior | Any:
        """Return the parameter of the prior."""
        return self.parameters[key]

    def _checks(self) -> None:
        if not self.centered:
            self._correct_non_centered_distribution()

        self._parameters_are_at_least_subset_of_pymc()
        self._convert_lists_to_numpy()
        self._parameters_are_correct_type()

    def _parameters_are_at_least_subset_of_pymc(self) -> None:
        pymc_params = _get_pymc_parameters(self.pymc_distribution)
        if not set(self.parameters.keys()).issubset(pymc_params):
            msg = (
                f"Parameters {set(self.parameters.keys())} "
                "are not a subset of the pymc distribution "
                f"parameters {set(pymc_params)}"
            )
            raise ValueError(msg)

    def _convert_lists_to_numpy(self) -> None:
        def convert(x):
            if not isinstance(x, list):
                return x

            return np.array(x)

        self.parameters = {key: convert(value) for key, value in self.parameters.items()}

    def _parameters_are_correct_type(self) -> None:
        supported_types = (
            int,
            float,
            np.ndarray,
            Prior,
            pt.TensorVariable,
            VariableFactory,
        )

        incorrect_types = {
            param: type(value)
            for param, value in self.parameters.items()
            if not isinstance(value, supported_types)
        }
        if incorrect_types:
            msg = (
                "Parameters must be one of the following types: "
                f"(int, float, np.array, Prior, pt.TensorVariable). Incorrect parameters: {incorrect_types}"
            )
            raise ValueError(msg)

    def _correct_non_centered_distribution(self) -> None:
        if not self.centered and self.distribution not in self.non_centered_distributions:
            raise UnsupportedParameterizationError(
                f"{self.distribution!r} is not supported for non-centered parameterization. "
                f"Choose from {list(self.non_centered_distributions.keys())}"
            )

        required_parameters = set(self.non_centered_distributions[self.distribution].keys())

        if set(self.parameters.keys()) < required_parameters:
            msg = " and ".join([f"{param!r}" for param in required_parameters])
            raise ValueError(
                f"Must have at least {msg} parameter for non-centered for {self.distribution!r}"
            )

    def _unique_dims(self) -> None:
        if not self.dims:
            return

        if len(self.dims) != len(set(self.dims)):
            raise ValueError("Dims must be unique")

    def _param_dims_work(self) -> None:
        other_dims = set()
        for value in self.parameters.values():
            if hasattr(value, "dims"):
                other_dims.update(value.dims)

        if not other_dims.issubset(self.dims):
            raise UnsupportedShapeError(
                f"Parameter dims {other_dims} are not a subset of the prior dims {self.dims}"
            )

    def __str__(self) -> str:
        """Return a string representation of the prior."""
        param_str = ", ".join([f"{param}={value}" for param, value in self.parameters.items()])
        param_str = "" if not param_str else f", {param_str}"

        dim_str = f", dims={_dims_to_str(self.dims)}" if self.dims else ""
        centered_str = f", centered={self.centered}" if not self.centered else ""
        transform_str = f', transform="{self.transform}"' if self.transform else ""
        return f'Prior("{self.distribution}"{param_str}{dim_str}{centered_str}{transform_str})'

    def __repr__(self) -> str:
        """Return a string representation of the prior."""
        return f"{self}"

    def _create_parameter(self, param, value, name):
        if not hasattr(value, "create_variable"):
            return value

        child_name = f"{name}_{param}"
        return self.dim_handler(value.create_variable(child_name), value.dims)

    def _create_centered_variable(self, name: str):
        parameters = {
            param: self._create_parameter(param, value, name)
            for param, value in self.parameters.items()
        }
        return self.pymc_distribution(name, **parameters, dims=self.dims)

    def _create_non_centered_variable(self, name: str) -> pt.TensorVariable:
        def handle_variable(var_name: str):
            parameter = self.parameters[var_name]
            if not hasattr(parameter, "create_variable"):
                return parameter

            return self.dim_handler(
                parameter.create_variable(f"{name}_{var_name}"),
                parameter.dims,
            )

        defaults = self.non_centered_distributions[self.distribution]
        other_parameters = {
            param: handle_variable(param)
            for param in self.parameters.keys()
            if param not in defaults
        }
        offset = self.pymc_distribution(
            f"{name}_offset",
            **defaults,
            **other_parameters,
            dims=self.dims,
        )
        if "mu" in self.parameters:
            mu = (
                handle_variable("mu")
                if isinstance(self.parameters["mu"], Prior)
                else self.parameters["mu"]
            )
        else:
            mu = 0

        sigma = (
            handle_variable("sigma")
            if isinstance(self.parameters["sigma"], Prior)
            else self.parameters["sigma"]
        )

        return pm.Deterministic(
            name,
            mu + sigma * offset,
            dims=self.dims,
        )

    def create_variable(self, name: str) -> pt.TensorVariable:
        """Create a PyMC variable from the prior.

        Must be used in a PyMC model context.

        Parameters
        ----------
        name : str
            The name of the variable.

        Returns
        -------
        pt.TensorVariable
            The PyMC variable.

        Examples
        --------
        Create a hierarchical normal variable in larger PyMC model.

        .. code-block:: python

            dist = Prior(
                "Normal",
                mu=Prior("Normal"),
                sigma=Prior("HalfNormal"),
                dims="channel",
            )

            coords = {"channel": ["C1", "C2", "C3"]}
            with pm.Model(coords=coords):
                var = dist.create_variable("var")

        """
        self.dim_handler = create_dim_handler(self.dims)

        if self.transform:
            var_name = f"{name}_raw"

            def transform(var):
                return pm.Deterministic(name, self.pytensor_transform(var), dims=self.dims)
        else:
            var_name = name

            def transform(var):
                return var

        create_variable = (
            self._create_centered_variable if self.centered else self._create_non_centered_variable
        )
        var = create_variable(name=var_name)
        return transform(var)

    @property
    def preliz(self):
        """Create an equivalent preliz distribution.

        Helpful to visualize a distribution when it is univariate.

        Returns
        -------
        preliz.distributions.Distribution

        Examples
        --------
        Create a preliz distribution from a prior.

        .. code-block:: python

            from pymc_extras.prior import Prior

            dist = Prior("Gamma", alpha=5, beta=1)
            dist.preliz.plot_pdf()

        """
        import preliz as pz

        return getattr(pz, self.distribution)(**self.parameters)

    def to_dict(self) -> dict[str, Any]:
        """Convert the prior to dictionary format.

        Returns
        -------
        dict[str, Any]
            The dictionary format of the prior.

        Examples
        --------
        Convert a prior to the dictionary format.

        .. code-block:: python

            from pymc_extras.prior import Prior

            dist = Prior("Normal", mu=0, sigma=1)

            dist.to_dict()

        Convert a hierarchical prior to the dictionary format.

        .. code-block:: python

            dist = Prior(
                "Normal",
                mu=Prior("Normal"),
                sigma=Prior("HalfNormal"),
                dims="channel",
            )

            dist.to_dict()

        """
        data: dict[str, Any] = {
            "dist": self.distribution,
        }
        if self.parameters:

            def handle_value(value):
                if isinstance(value, Prior):
                    return value.to_dict()

                if isinstance(value, pt.TensorVariable):
                    value = value.eval()

                if isinstance(value, np.ndarray):
                    return value.tolist()

                if hasattr(value, "to_dict"):
                    return value.to_dict()

                return value

            data["kwargs"] = {
                param: handle_value(value) for param, value in self.parameters.items()
            }
        if not self.centered:
            data["centered"] = False

        if self.dims:
            data["dims"] = self.dims

        if self.transform:
            data["transform"] = self.transform

        return data

    @classmethod
    def from_dict(cls, data) -> Prior:
        """Create a Prior from the dictionary format.

        Parameters
        ----------
        data : dict[str, Any]
            The dictionary format of the prior.

        Returns
        -------
        Prior
            The prior distribution.

        Examples
        --------
        Convert prior in the dictionary format to a Prior instance.

        .. code-block:: python

            from pymc_extras.prior import Prior

            data = {
                "dist": "Normal",
                "kwargs": {"mu": 0, "sigma": 1},
            }

            dist = Prior.from_dict(data)
            dist
            # Prior("Normal", mu=0, sigma=1)

        """
        if not isinstance(data, dict):
            msg = (
                "Must be a dictionary representation of a prior distribution. "
                f"Not of type: {type(data)}"
            )
            raise ValueError(msg)

        dist = data["dist"]
        kwargs = data.get("kwargs", {})

        def handle_value(value):
            if isinstance(value, dict):
                return deserialize(value)

            if isinstance(value, list):
                return np.array(value)

            return value

        kwargs = {param: handle_value(value) for param, value in kwargs.items()}
        centered = data.get("centered", True)
        dims = data.get("dims")
        if isinstance(dims, list):
            dims = tuple(dims)
        transform = data.get("transform")

        return cls(dist, dims=dims, centered=centered, transform=transform, **kwargs)

    def constrain(self, lower: float, upper: float, mass: float = 0.95, kwargs=None) -> Prior:
        """Create a new prior with a given mass constrained within the given bounds.

        Wrapper around `preliz.maxent`.

        Parameters
        ----------
        lower : float
            The lower bound.
        upper : float
            The upper bound.
        mass: float = 0.95
            The mass of the distribution to keep within the bounds.
        kwargs : dict
            Additional arguments to pass to `pz.maxent`.

        Returns
        -------
        Prior
            The maximum entropy prior with a mass constrained to the given bounds.

        Examples
        --------
        Create a Beta distribution that is constrained to have 95% of the mass
        between 0.5 and 0.8.

        .. code-block:: python

            dist = Prior(
                "Beta",
            ).constrain(lower=0.5, upper=0.8)

        Create a Beta distribution with mean 0.6, that is constrained to
        have 95% of the mass between 0.5 and 0.8.

        .. code-block:: python

            dist = Prior(
                "Beta",
                mu=0.6,
            ).constrain(lower=0.5, upper=0.8)

        """
        from preliz import maxent

        if self.transform:
            raise ValueError("Can't constrain a transformed variable")

        if kwargs is None:
            kwargs = {}
            kwargs.setdefault("plot", False)

        if kwargs["plot"]:
            new_parameters = maxent(self.preliz, lower, upper, mass, **kwargs)[0].params_dict
        else:
            new_parameters = maxent(self.preliz, lower, upper, mass, **kwargs).params_dict

        return Prior(
            self.distribution,
            dims=self.dims,
            transform=self.transform,
            centered=self.centered,
            **new_parameters,
        )

    def __eq__(self, other) -> bool:
        """Check if two priors are equal."""
        if not isinstance(other, Prior):
            return False

        try:
            np.testing.assert_equal(self.parameters, other.parameters)
        except AssertionError:
            return False

        return (
            self.distribution == other.distribution
            and self.dims == other.dims
            and self.centered == other.centered
            and self.transform == other.transform
        )

    def sample_prior(
        self,
        coords=None,
        name: str = "variable",
        **sample_prior_predictive_kwargs,
    ) -> xr.Dataset:
        """Sample the prior distribution for the variable.

        Parameters
        ----------
        coords : dict[str, list[str]], optional
            The coordinates for the variable, by default None.
            Only required if the dims are specified.
        name : str, optional
            The name of the variable, by default "variable".
        sample_prior_predictive_kwargs : dict
            Additional arguments to pass to `pm.sample_prior_predictive`.

        Returns
        -------
        xr.Dataset
            The dataset of the prior samples.

        Example
        -------
        Sample from a hierarchical normal distribution.

        .. code-block:: python

            dist = Prior(
                "Normal",
                mu=Prior("Normal"),
                sigma=Prior("HalfNormal"),
                dims="channel",
            )

            coords = {"channel": ["C1", "C2", "C3"]}
            prior = dist.sample_prior(coords=coords)

        """
        return sample_prior(
            factory=self,
            coords=coords,
            name=name,
            **sample_prior_predictive_kwargs,
        )

    def __deepcopy__(self, memo) -> Prior:
        """Return a deep copy of the prior."""
        if id(self) in memo:
            return memo[id(self)]

        copy_obj = Prior(
            self.distribution,
            dims=copy.copy(self.dims),
            centered=self.centered,
            transform=self.transform,
            **copy.deepcopy(self.parameters),
        )
        memo[id(self)] = copy_obj
        return copy_obj

    def deepcopy(self) -> Prior:
        """Return a deep copy of the prior."""
        return copy.deepcopy(self)

    def to_graph(self):
        """Generate a graph of the variables.

        Examples
        --------
        Create the graph for a 2D transformed hierarchical distribution.

        .. code-block:: python

            from pymc_extras.prior import Prior

            mu = Prior(
                "Normal",
                mu=Prior("Normal"),
                sigma=Prior("HalfNormal"),
                dims="channel",
            )
            sigma = Prior("HalfNormal", dims="channel")
            dist = Prior(
                "Normal",
                mu=mu,
                sigma=sigma,
                dims=("channel", "geo"),
                centered=False,
                transform="sigmoid",
            )

            dist.to_graph()

        .. image:: /_static/example-graph.png
            :alt: Example graph

        """
        coords = {name: ["DUMMY"] for name in self.dims}
        with pm.Model(coords=coords) as model:
            self.create_variable("var")

        return pm.model_to_graphviz(model)

    def create_likelihood_variable(
        self,
        name: str,
        mu: pt.TensorLike,
        observed: pt.TensorLike,
    ) -> pt.TensorVariable:
        """Create a likelihood variable from the prior.

        Will require that the distribution has a `mu` parameter
        and that it has not been set in the parameters.

        Parameters
        ----------
        name : str
            The name of the variable.
        mu : pt.TensorLike
            The mu parameter for the likelihood.
        observed : pt.TensorLike
            The observed data.

        Returns
        -------
        pt.TensorVariable
            The PyMC variable.

        Examples
        --------
        Create a likelihood variable in a larger PyMC model.

        .. code-block:: python

            import pymc as pm

            dist = Prior("Normal", sigma=Prior("HalfNormal"))

            with pm.Model():
                # Create the likelihood variable
                mu = pm.Normal("mu", mu=0, sigma=1)
                dist.create_likelihood_variable("y", mu=mu, observed=observed)

        """
        if "mu" not in _get_pymc_parameters(self.pymc_distribution):
            raise UnsupportedDistributionError(
                f"Likelihood distribution {self.distribution!r} is not supported."
            )

        if "mu" in self.parameters:
            raise MuAlreadyExistsError(self)

        distribution = self.deepcopy()
        distribution.parameters["mu"] = mu
        distribution.parameters["observed"] = observed
        return distribution.create_variable(name)


class VariableNotFound(Exception):
    """Variable is not found."""


def _remove_random_variable(var: pt.TensorVariable) -> None:
    if var.name is None:
        raise ValueError("This isn't removable")

    name: str = var.name

    model = pm.modelcontext(None)
    for idx, free_rv in enumerate(model.free_RVs):
        if var == free_rv:
            index_to_remove = idx
            break
    else:
        raise VariableNotFound(f"Variable {var.name!r} not found")

    var.name = None
    model.free_RVs.pop(index_to_remove)
    model.named_vars.pop(name)


@dataclass
class Censored:
    """Create censored random variable.

    Examples
    --------
    Create a censored Normal distribution:

    .. code-block:: python

        from pymc_extras.prior import Prior, Censored

        normal = Prior("Normal")
        censored_normal = Censored(normal, lower=0)

    Create hierarchical censored Normal distribution:

    .. code-block:: python

        from pymc_extras.prior import Prior, Censored

        normal = Prior(
            "Normal",
            mu=Prior("Normal"),
            sigma=Prior("HalfNormal"),
            dims="channel",
        )
        censored_normal = Censored(normal, lower=0)

        coords = {"channel": range(3)}
        samples = censored_normal.sample_prior(coords=coords)

    """

    distribution: InstanceOf[Prior]
    lower: float | InstanceOf[pt.TensorVariable] = -np.inf
    upper: float | InstanceOf[pt.TensorVariable] = np.inf

    def __post_init__(self) -> None:
        """Check validity at initialization."""
        if not self.distribution.centered:
            raise ValueError(
                "Censored distribution must be centered so that .dist() API can be used on distribution."
            )

        if self.distribution.transform is not None:
            raise ValueError(
                "Censored distribution can't have a transform so that .dist() API can be used on distribution."
            )

    @property
    def dims(self) -> tuple[str, ...]:
        """The dims from the distribution to censor."""
        return self.distribution.dims

    @dims.setter
    def dims(self, dims) -> None:
        self.distribution.dims = dims

    def create_variable(self, name: str) -> pt.TensorVariable:
        """Create censored random variable."""
        dist = self.distribution.create_variable(name)
        _remove_random_variable(var=dist)

        return pm.Censored(
            name,
            dist,
            lower=self.lower,
            upper=self.upper,
            dims=self.dims,
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert the censored distribution to a dictionary."""

        def handle_value(value):
            if isinstance(value, pt.TensorVariable):
                return value.eval().tolist()

            return value

        return {
            "class": "Censored",
            "data": {
                "dist": self.distribution.to_dict(),
                "lower": handle_value(self.lower),
                "upper": handle_value(self.upper),
            },
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Censored:
        """Create a censored distribution from a dictionary."""
        data = data["data"]
        return cls(  # type: ignore
            distribution=deserialize(data["dist"]),
            lower=data["lower"],
            upper=data["upper"],
        )

    def sample_prior(
        self,
        coords=None,
        name: str = "variable",
        **sample_prior_predictive_kwargs,
    ) -> xr.Dataset:
        """Sample the prior distribution for the variable.

        Parameters
        ----------
        coords : dict[str, list[str]], optional
            The coordinates for the variable, by default None.
            Only required if the dims are specified.
        name : str, optional
            The name of the variable, by default "var".
        sample_prior_predictive_kwargs : dict
            Additional arguments to pass to `pm.sample_prior_predictive`.

        Returns
        -------
        xr.Dataset
            The dataset of the prior samples.

        Example
        -------
        Sample from a censored Gamma distribution.

        .. code-block:: python

            gamma = Prior("Gamma", mu=1, sigma=1, dims="channel")
            dist = Censored(gamma, lower=0.5)

            coords = {"channel": ["C1", "C2", "C3"]}
            prior = dist.sample_prior(coords=coords)

        """
        return sample_prior(
            factory=self,
            coords=coords,
            name=name,
            **sample_prior_predictive_kwargs,
        )

    def to_graph(self):
        """Generate a graph of the variables.

        Examples
        --------
        Create graph for a censored Normal distribution

        .. code-block:: python

            from pymc_extras.prior import Prior, Censored

            normal = Prior("Normal")
            censored_normal = Censored(normal, lower=0)

            censored_normal.to_graph()

        """
        coords = {name: ["DUMMY"] for name in self.dims}
        with pm.Model(coords=coords) as model:
            self.create_variable("var")

        return pm.model_to_graphviz(model)

    def create_likelihood_variable(
        self,
        name: str,
        mu: pt.TensorLike,
        observed: pt.TensorLike,
    ) -> pt.TensorVariable:
        """Create observed censored variable.

        Will require that the distribution has a `mu` parameter
        and that it has not been set in the parameters.

        Parameters
        ----------
        name : str
            The name of the variable.
        mu : pt.TensorLike
            The mu parameter for the likelihood.
        observed : pt.TensorLike
            The observed data.

        Returns
        -------
        pt.TensorVariable
            The PyMC variable.

        Examples
        --------
        Create a censored likelihood variable in a larger PyMC model.

        .. code-block:: python

            import pymc as pm
            from pymc_extras.prior import Prior, Censored

            normal = Prior("Normal", sigma=Prior("HalfNormal"))
            dist = Censored(normal, lower=0)

            observed = 1

            with pm.Model():
                # Create the likelihood variable
                mu = pm.HalfNormal("mu", sigma=1)
                dist.create_likelihood_variable("y", mu=mu, observed=observed)

        """
        if "mu" not in _get_pymc_parameters(self.distribution.pymc_distribution):
            raise UnsupportedDistributionError(
                f"Likelihood distribution {self.distribution.distribution!r} is not supported."
            )

        if "mu" in self.distribution.parameters:
            raise MuAlreadyExistsError(self.distribution)

        distribution = self.distribution.deepcopy()
        distribution.parameters["mu"] = mu

        dist = distribution.create_variable(name)
        _remove_random_variable(var=dist)

        return pm.Censored(
            name,
            dist,
            observed=observed,
            lower=self.lower,
            upper=self.upper,
            dims=self.dims,
        )


class Scaled:
    """Scaled distribution for numerical stability."""

    def __init__(self, dist: Prior, factor: float | pt.TensorVariable) -> None:
        self.dist = dist
        self.factor = factor

    @property
    def dims(self) -> Dims:
        """The dimensions of the scaled distribution."""
        return self.dist.dims

    def create_variable(self, name: str) -> pt.TensorVariable:
        """Create a scaled variable.

        Parameters
        ----------
        name : str
            The name of the variable.

        Returns
        -------
        pt.TensorVariable
            The scaled variable.
        """
        var = self.dist.create_variable(f"{name}_unscaled")
        return pm.Deterministic(name, var * self.factor, dims=self.dims)


def _is_prior_type(data: dict) -> bool:
    return "dist" in data


def _is_censored_type(data: dict) -> bool:
    return data.keys() == {"class", "data"} and data["class"] == "Censored"


register_deserialization(is_type=_is_prior_type, deserialize=Prior.from_dict)
register_deserialization(is_type=_is_censored_type, deserialize=Censored.from_dict)


def __getattr__(name: str):
    """Get Prior class through the module.

    Examples
    --------
    Create a normal distribution.

    .. code-block:: python

        from pymc_extras.prior import Normal

        dist = Normal(mu=1, sigma=2)

    Create a hierarchical normal distribution.

    .. code-block:: python

        import pymc_extras.prior as pr

        dist = pr.Normal(mu=pr.Normal(), sigma=pr.HalfNormal(), dims="channel")
        samples = dist.sample_prior(coords={"channel": ["C1", "C2", "C3"]})

    """
    # Protect against doctest
    if name == "__wrapped__":
        return

    _get_pymc_distribution(name)
    return partial(Prior, distribution=name)



from functools import wraps
from inspect import signature

import pytensor.tensor as pt

from pymc import Data, Model


def as_model(*model_args, **model_kwargs):
    R"""
    Decorator to provide context to PyMC models declared in a function.
    This removes all need to think about context managers and lets you separate creating a generative model from using the model.
    Additionally, a coords argument is added to the function so coords can be changed during function invocation

    All parameters are wrapped with a `pm.Data` object if the underlying type of the data supports it.

    Adapted from `Rob Zinkov's blog post <https://www.zinkov.com/posts/2023-alternative-frontends-pymc/>`_ and inspired by the `sampled <https://github.com/colcarroll/sampled>`_ decorator for PyMC3.

    Examples
    --------
    .. code:: python

        import pymc as pm
        import pymc_extras as pmx

        # The following are equivalent

        # standard PyMC API with context manager
        with pm.Model(coords={"obs": ["a", "b"]}) as model:
            x = pm.Normal("x", 0., 1., dims="obs")
            pm.sample()

        # functional API using decorator
        @pmx.as_model(coords={"obs": ["a", "b"]})
        def basic_model():
            pm.Normal("x", 0., 1., dims="obs")

        m = basic_model()
        pm.sample(model=m)

        # alternative way to use functional API
        @pmx.as_model()
        def basic_model():
            pm.Normal("x", 0., 1., dims="obs")

        m = basic_model(coords={"obs": ["a", "b"]})
        pm.sample(model=m)

    """

    def decorator(f):
        @wraps(f)
        def make_model(*args, **kwargs):
            coords = model_kwargs.pop("coords", {}) | kwargs.pop("coords", {})
            sig = signature(f)
            ba = sig.bind(*args, **kwargs)
            ba.apply_defaults()

            with Model(*model_args, coords=coords, **model_kwargs) as m:
                for name, v in ba.arguments.items():
                    # Only wrap pm.Data around values pytensor can process
                    try:
                        _ = pt.as_tensor_variable(v)
                        ba.arguments[name] = Data(name, v)
                    except (NotImplementedError, TypeError, ValueError):
                        pass
                f(*ba.args, **ba.kwargs)
            return m

        return make_model

    return decorator



from typing import cast as type_cast

import numpy as np
import pytensor.tensor as pt

from pytensor.tensor import TensorVariable

from pymc_extras.statespace.utils.constants import (
    ALL_STATE_AUX_DIM,
    ALL_STATE_DIM,
    LONG_MATRIX_NAMES,
    MATRIX_NAMES,
    OBS_STATE_AUX_DIM,
    OBS_STATE_DIM,
    SHOCK_AUX_DIM,
    SHOCK_DIM,
    VECTOR_VALUED,
)


def make_default_coords(ss_mod):
    coords = {
        ALL_STATE_DIM: ss_mod.state_names,
        ALL_STATE_AUX_DIM: ss_mod.state_names,
        OBS_STATE_DIM: ss_mod.observed_states,
        OBS_STATE_AUX_DIM: ss_mod.observed_states,
        SHOCK_DIM: ss_mod.shock_names,
        SHOCK_AUX_DIM: ss_mod.shock_names,
    }

    return coords


def cleanup_states(states: list[str]) -> list[str]:
    """
    Remove meaningless symbols from state names

    Parameters
    ----------
    states, list of str
        State names generated by make_harvey_state_names

    Returns
    -------
    states, list of str
        State names for the Harvey statespace representation, with meaningless terms removed

    The state names generated by make_harvey_state_names includes some "meaningless" terms. For example, lags are
    indicated with L{i}.state. This includes L0.state, which is correctly just "state".

    In addition, sequential applications of the difference operator are denoted Dk^i, where k is the length of the
    difference, and i is the number of repeated applications. Dk^1 is thus just Dk.
    """

    out = []
    for state in states:
        state = state.replace("^1", "")
        state = state.replace("^0", "")
        state = state.replace("L0", "")
        state = state.replace("D0", "")
        out.append(state)
    return out


def make_harvey_state_names(p: int, d: int, q: int, P: int, D: int, Q: int, S: int) -> list[str]:
    """
    Generate informative names for the SARIMA states in the Harvey representation

    Parameters
    ----------
    p: int
        AR order
    d: int
        Number of ARIMA differences
    q: int
        MA order
    P: int
        Seasonal AR order
    D: int
        Number of seasonal differences
    Q: int
        Seasonal MA order
    S: int
        Seasonal length

    Returns
    -------
    state_names, list of str
        List of state names

    The Harvey state is not particularly interpretable, but it's also not totally opaque. This helper function makes
    a list of state names that can help users understand what they are getting back from the statespace. In particular,
    it is helpful to know how differences and seasonal differences are incorporated into the model
    """
    k_lags = max(p + P * S, q + Q * S + 1)
    has_diff = (d + D) > 0

    # First state is always data
    states = ["data"]

    # Differencing operations
    # The goal here is to get down to "data_star", the state that actually has the SARIMA dynamics applied to it.
    # To get there, first the data needs to be differenced d-1 times
    d_size = d + int(D > 0)
    states.extend([f"D1^{(i + 1)}.data" for i in range(d_size)[:-1]])

    # Next, if there are seasonal differences, we need to lag the ARIMA differenced state S times, then seasonal
    # difference it. This procedure is done D-1 times.

    arma_diff = [int(d_size > 1), d_size - 1]
    season_diff = [S, 0]
    curr_state = f"D{arma_diff[0]}^{arma_diff[1]}"
    for i in range(D):
        states.extend([f"L{j + 1}{curr_state}.data" for j in range(S - 1)])
        season_diff[1] += 1
        curr_state = f"D{arma_diff[0]}^{arma_diff[1]}D{season_diff[0]}^{season_diff[1]}"
        if i != (D - 1):
            states.append(f"{curr_state}.data")

    # Now we are at data_star. If we did any differencing, add it in.
    if has_diff:
        states.append("data_star")

    # Next, we add the time series dynamics states. These don't have a immediately obvious interpretation, so just call
    # them "state_1" .., "state_n".
    suffix = "_star" if "star" in states[-1] else ""
    states.extend([f"state{suffix}_{i + 1}" for i in range(k_lags - 1)])

    states = cleanup_states(states)

    return states


def make_SARIMA_transition_matrix(
    p: int, d: int, q: int, P: int, D: int, Q: int, S: int
) -> np.ndarray:
    r"""
    Make the transition matrix for a SARIMA model

    Parameters
    ----------
    p: int
        AR order
    d: int
        Number of ARIMA differences
    q: int
        MA order
    P: int
        Seasonal AR order
    D: int
        Number of seasonal differences
    Q: int
        Seasonal MA order
    S: int
        Seasonal length

    Returns
    -------
    T, ndarray
        The transition matrix associated with a SARIMA model of order (p,d,q)x(P,D,Q,S)

    Notes
    -----
    The transition matrix for the SARIMA model has a bunch of structure in it, especially when differences are included
    in the statespace model. This function will always assume the state space matrix is in the Harvey representation.

    Given this representation, the matrix can be divided into a bottom part and a top part. The top part has (S * D) + d
    rows, and is associated with the differencing operations. The bottom part has max(P*S+p, Q*S+q+1) rows, and is
    responsible for the actual time series dynamics.

    The bottom part of the matrix is quite simple, it is just a shifted identity matrix (called a "companion matrix"),
    responsible for "rolling" the states, so that at each transition, the value for :math:`x_{t-3}` becomes the value
    for :math:`x_{t-2}`, and so on.

    The top part is quite complex. The goal of this part of the matrix is to transform the raw data state, :math:`x_t`,
    into a stationary state, :math:`x_t^\star`, via the application of differencing operations,
    :math:`\Delta x_t = x_t - x_{t-1}`. For ARIMA differences (the little ``d``), this is quite simple. Sequential
    differences are representated as an upper-triangular matrix of ones. To see this, consider an example where ``d=3``,
    so that:
     .. math::

        \begin{align}
            x_t^\star &= \Delta^3 x_t \\
             &= \Delta^2 (x_t - x_{t-1})
             &= \Delta (x_t - 2x_{t-1} + x_{t-2})
             &= x_t - x_{t-1} - 2x_{t-1} + 2x_{t-3} + x_{t-2} - x_{t-3}
             &= x_t - 3x_{t-1} + 3x_{t-3} - x_{t-3}
        \end{align}

    If you choose a state vector :math:`\begin{bmatrix}x_t & \Delta x_t & \Delta^2 x_t & x_t^\star \end{bmatrix}^T`,
    you will find that:

    .. math::
        \begin{bmatrix}x_t \\ \Delta x_t \\ \Delta^2 x_t \\ x_t^\star \end{bmatrix} =
            \begin{bmatrix} 1 & 1 & 1 & 1 \\
                            0 & 1 & 1 & 1 \\
                            0 & 0 & 1 & 1 \\
                            0 & 0 & 0 & 1
            \end{bmatrix}
            \begin{bmatrix} x_{t-1} \\ \Delta x_{t-1} \\ \Delta^2 x_{t-1} \\ x_{t-1}^\star \end{bmatrix}

    Next are the seasonal differences. The highest seasonal difference stored in the states is one less than the
    seasonal difference order, ``D``. That is, if ``D = 1, S = 4``, there will be states :math:``x_{t-1}, x_{t-2},
    x_{t-3}, x_{t-4}, x_t^\star`, with :math:`x_t^\star = \Delta_4 x_t = x_t - x_{t-4}`. The level state can be
    recovered by adding :math:`x_t^\star + x_{t-4}`. To accomplish all of this, two things need to be inserted into the
    transition matrix:

        1. A shifted identity matrix to "roll" the lagged states forward each transition, and
        2. A pair of 1's to recover the level state by adding the last 2 states (:math:`x_t^\star + x_{t-4}`)

    Keeping the example of ``D = 1, S = 4``, the block that handles the seasonal difference will look this this:
    .. math::
        \begin{bmatrix} 0 & 0 & 0 & 1 & 1 \\
                        1 & 0 & 0 & 0 & 0 \\
                        0 & 1 & 0 & 0 & 0 \\
                        0 & 0 & 1 & 0 & 0 \\
                        0 & 0 & 0 & 0 & 0 \end{bmatrix}

    In the presence of higher order seasonal differences, there needs to be one block per difference. And the level
    state is recovered by adding together the last state from each block. For example, if ``D = 2, S = 4``, the states
    will be :math:`x_{t-1}, x_{t-2}, x_{t-3}, x_{t-4}, \Delta_4 x_{t-1}, \Delta_4 x_{t-2}, \Delta_4 x_{t-3},
    \Delta_4 x_{t-4} x_t^\star`, with :math:`x_t^\star = \Delta_4^2 = \Delta_4(x_t - x_{t-4}) = x_t - 2 x_{t-4} +
    x_{t-8}`. To recover the level state, we need :math:`x_t = x_t^\star + \Delta_4 x_{t-4} + x_{t-4}`. In addition,
    to recover :math:`\Delta_4 x_t`, we have to compute :math:`\Delta_4 x_t = x_t^\star + \Delta_4 x_{t-4} =
    \Delta_4(x_t - x_{t-4}) + \Delta_4 x_{t-4} = \Delta_4 x_t`. The block of the transition matrix associated with all
    this is thus:

    .. math::
        \begin{bmatrix} 0 & 0 & 0 & 1 & 0 & 0 & 0 & 1 & 1 \\
                        1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                        0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                        0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                        0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 1 \\
                        0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                        0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                        0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                        0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \end{bmatrix}

    When ARIMA differences and seasonal differences are mixed, the seasonal differences will be written in terms of the
    highest ARIMA difference order, and recovery of the level state will require the use of all the ARIMA differences,
    as well as the seasonal differences. In addition, the seasonal differences are needed to back out the ARIMA
    differences from :math:`x_t^\star`. Here is the differencing block for a SARIMA(0,2,0)x(0,2,0,4) -- the identites
    of the states is left an exercise for the motivated reader:

    .. math::
        \begin{bmatrix}
            1 & 1 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 1 & 1 \\
            0 & 1 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 1 & 1 \\
            0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 1 & 1 \\
            0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 1 \\
            0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \end{bmatrix}
    """
    n_diffs = S * D + d
    k_lags = max(p + P * S, q + Q * S + 1)
    k_states = k_lags + n_diffs

    # Top Part
    # ARIMA differences
    T = np.zeros((k_states, k_states))
    diff_idx = np.triu_indices(d)
    T[diff_idx] = 1

    # Adjustment factors for difference states All of the difference states are computed relative to x_t_star using
    # combinations of states, so there's a lot of "backing out" that needs to happen here. The columns are the more
    # straightforward part. After the (d,d) upper triangle of 1s for the ARIMA lags, there will be (S - 1) zeros,
    # and then a 1. In addition, there is an extra column of 1s at position n_diffs + 1, corresponding to x_star itself.

    # This will slowly taper down, but first we build the "full" set of column indices with values
    base_col_idx = d + S + np.arange(D) * S - 1
    if len(base_col_idx) > 0:
        base_col_idx = np.r_[base_col_idx, base_col_idx[-1] + 1]

    # The first d rows -- associated with the ARIMA differences -- will have 1s in all columns.
    col_idx = np.tile(base_col_idx, d)
    row_idx = np.arange(d).repeat(D + 1)

    # Next, if there are seasonal differences, there will be more rows, with the columns slowly dropping off.
    # Starting from the d+1-th row, there will be 1 in the column positions every S rows, for a total of (D-1) rows.
    # Every row will drop 2 columns from the left of base_col_idx.
    for i in range(D):
        n = len(base_col_idx[i:])
        col_idx = np.r_[col_idx, base_col_idx[i:]]
        row_idx = np.r_[row_idx, np.full(n, d + S * i)]

    if D == 0 and d > 0:
        # Special case: If there are *only* ARIMA lags, there still needs to be a single column of 1s at position
        # [:d, d]
        row_idx = np.arange(d)
        col_idx = np.full(d, d)
    T[row_idx, col_idx] = 1

    if S > 0:
        # "Rolling" indices for seasonal differences
        (row_roll_idx, col_roll_idx) = np.diag_indices(S * D)
        row_roll_idx = row_roll_idx + d + 1
        col_roll_idx = col_roll_idx + d

        # Rolling indices have a zero after every diagonal of length S-1
        T[row_roll_idx, col_roll_idx] = 1
        zero_idx = row_roll_idx[S - 1 :: S], col_roll_idx[S - 1 :: S]
        T[zero_idx] = 0

    # Bottom part
    # Rolling indices for the "compute" states, x_star
    star_roll_row, star_roll_col = np.diag_indices(k_lags - 1)
    star_roll_row = star_roll_row + n_diffs
    star_roll_col = star_roll_col + n_diffs + 1

    T[star_roll_row, star_roll_col] = 1

    return T


def conform_time_varying_and_time_invariant_matrices(A, B):
    """
    Adjust either A or B to conform to the other in the time dimension

    In the context of building a structural model from components, it might be the case that one component has
    time-varying statespace matrices, while the other does not. In this case, it is not possible to concatenate
    or block diagonalize the pair of matrices A and B without first expanding the time-invariant matrix to have a
    time dimension. This function checks if exactly one of the two time varies, and adjusts the other accordingly if
    need be.

    Parameters
    ----------
    A: pt.TensorVariable
        An anonymous statespace matrix
    B: pt.TensorVariable
        An anonymous statespace matrix

    Returns
    -------
    (A, B): Tuple of pt.TensorVariable
        A and B, with one or neither expanded to have a time dimension.
    """

    if A.name == B.name:
        name = A.name
    else:
        if all([X.name not in MATRIX_NAMES + LONG_MATRIX_NAMES for X in [A, B]]):
            raise ValueError(
                "At least one matrix passed to conform_time_varying_and_time_invariant_matrices should be a "
                "statespace matrix"
            )
        name = A.name if A.name in MATRIX_NAMES + LONG_MATRIX_NAMES else B.name

    time_varying_ndim = 3 - int(name in VECTOR_VALUED)

    if not all([x.ndim == time_varying_ndim for x in [A, B]]):
        return A, B

    T_A, *A_dims = A.type.shape
    T_B, *B_dims = B.type.shape

    if T_A == T_B:
        return A, B

    if T_A == 1:
        A_out = pt.repeat(A, B.shape[0], axis=0)
        A_out = pt.specify_shape(A_out, (T_B, *tuple(A_dims)))
        A_out.name = A.name

        return A_out, B

    if T_B == 1:
        B_out = pt.repeat(B, A.shape[0], axis=0)
        B_out = pt.specify_shape(B_out, (T_A, *tuple(B_dims)))
        B_out.name = B.name

        return A, B_out

    return A, B


def normalize_axis(x, axis):
    """
    Convert negative axis values to positive axis values
    """
    if isinstance(axis, tuple):
        return tuple([normalize_axis(x, i) for i in axis])
    if axis < 0:
        axis = x.ndim + axis
    return axis


def reorder_from_labels(
    x: TensorVariable,
    labels: list[str],
    ordered_labels: list[str],
    labeled_axis: int | tuple[int, int],
) -> TensorVariable:
    """
    Reorder an input tensor along request axis/axes based on lists of string labels

    Parameters
    ----------
    x: TensorVariable
        Input tensor
    labels: list of str
        Labels associated with values of the input tensor ``x``, along the ``labeled_axis``. At runtime, should have
        ``x.shape[labeled_axis] == len(labels)``
    ordered_labels: list of str
        Target ordering according to which ``x`` will be reordered.
    labeled_axis: int or tuple of int
        Axis along which ``x`` will be labeled. If a tuple, each axis will be assumed to have identical labels, and
        and reorganization will be done on all requested axes together (NOT fancy indexing!)

    Returns
    -------
    x_sorted: TensorVariable
        Output tensor sorted along ``labeled_axis`` according to ``ordered_labels``
    """
    n_out = len(ordered_labels)
    label_to_index = {label: index for index, label in enumerate(ordered_labels)}

    missing_labels = [label for label in ordered_labels if label not in labels]
    indices = np.argsort([label_to_index[label] for label in [*labels, *missing_labels]])

    if isinstance(labeled_axis, int):
        labeled_axis = (labeled_axis,)

    if indices.tolist() != list(range(n_out)):
        for axis in labeled_axis:
            idx = np.s_[tuple([slice(None, None) if i != axis else indices for i in range(x.ndim)])]
            shape = x.type.shape
            x = pt.specify_shape(x[idx], shape)

    return x


def pad_and_reorder(
    x: TensorVariable, labels: list[str], ordered_labels: list[str], labeled_axis: int
) -> TensorVariable:
    """
    Pad input tensor ``x`` along the `labeled_axis` to match the length of ``ordered_labels``, then reorder the
    padded dimension to match the ordering in ``ordered_labels``.

    Parameters
    ----------
    x: TensorVariable
        Input tensor
    labels: list of str
        String labels associated with the `x` tensor at the ``labeled_axis`` dimension. At runtime, should have
        ``x.shape[labeled_axis] == len(labels)``. ``labels`` should be a subset of ``ordered_labels``.
    ordered_labels: list of str
        Target ordering according to which ``x`` will be reordered.
    labeled_axis: int
        Axis along which ``x`` will be labeled.

    Returns
    -------
    x_padded: TensorVariable
        Output tensor padded along ``labeled_axis`` according to ``ordered_labels``, then reordered.

    """
    n_out = len(ordered_labels)
    n_missing = n_out - len(labels)

    if n_missing > 0:
        zeros = pt.zeros(
            tuple([x.shape[i] if i != labeled_axis else n_missing for i in range(x.ndim)])
        )
        x_padded = pt.concatenate([x, zeros], axis=labeled_axis)
    else:
        x_padded = x

    return reorder_from_labels(x_padded, labels, ordered_labels, labeled_axis)


def ndim_pad_and_reorder(
    x: TensorVariable,
    labels: list[str],
    ordered_labels: list[str],
    labeled_axis: int | tuple[int, int],
) -> TensorVariable:
    """
    Pad input tensor ``x`` along the `labeled_axis` to match the length of ``ordered_labels``, then reorder the
    padded dimension to match the ordering in ``ordered_labels``.

    Unlike ``pad_and_reorder``, this function allows padding and reordering to be done simultaneously on multiple
    axes. In this case, reordering is done jointly on all axes -- it does *not* use fancy indexing.

    Parameters
    ----------
    x: TensorVariable
        Input tensor
    labels: list of str
        Labels associated with values of the input tensor ``x``, along the ``labeled_axis``. At runtime, should have
        ``x.shape[labeled_axis] == len(labels)``. If ``labeled_axis`` is a tuple, all axes are assumed to have the
        same labels.
    ordered_labels: list of str
        Target ordering according to which ``x`` will be reordered. ``labels`` should be a subset of ``ordered_labels``.
    labeled_axis: int or tuple of int
        Axis along which ``x`` will be labeled. If a tuple, each axis will be assumed to have identical labels, and
        and reorganization will be done on all requested axes together (NOT fancy indexing!)

    Returns
    -------
    x_sorted: TensorVariable
        Output tensor. Each ``labeled_axis`` is padded to the length of ``ordered_labels``, then reordered.
    """
    n_missing = len(ordered_labels) - len(labels)

    if isinstance(labeled_axis, int):
        labeled_axis = (labeled_axis,)

    if n_missing > 0:
        pad_size = [(0, 0) if i not in labeled_axis else (0, n_missing) for i in range(x.ndim)]
        for axis, (_, after) in enumerate(pad_size):
            if after > 0:
                shape = list(x.type.shape)
                shape[axis] = after
                zero_shape = [
                    static_shape if static_shape is not None else x.shape[i]
                    for i, static_shape in enumerate(shape)
                ]
                x = pt.join(axis, x, pt.zeros(zero_shape))

    return reorder_from_labels(x, labels, ordered_labels, labeled_axis)


def add_tensors_by_dim_labels(
    tensor: TensorVariable,
    other_tensor: TensorVariable,
    labels: list[str],
    other_labels: list[str],
    labeled_axis: int | tuple[int, int] = -1,
) -> TensorVariable:
    """
    Add two tensors based on labels associated with one dimension.

    When combining statespace matrices associated with structural components with potentially different states, it is
    important to make sure that duplicated states are handled correctly. For bias vectors and covariance matrices,
    duplicated states should be summed.

    When a state appears in one component but not another, that state should be treated as an implicit zero in the
    components where the state does not appear. This amounts to padding the relevant matrices with zeros before
    performing the addition.

    When labeled_axis is a tuple, each provided label is assumed to be identically labeled in each input tensor. This
    is the case, for example, when working with a covariance matrix. In this case, padding and alignment will be
    done on each indicated index.

    Parameters
    ----------
    tensor: TensorVariable
        A statespace matrix to be summed with ``other_tensor``.
    other_tensor: TensorVariable
        A statespace matrix to be summed with ``tensor``.
    labels: list of str
        Dimension labels associated with ``tensor``, on the ``labeled_axis`` dimension.
    other_labels: list of str
        Dimension labels associated with ``other_tensor``, on the ``labeled_axis`` dimension.
    labeled_axis: int or tuple of int
        Dimension that is labeled by ``labels`` and ``other_labels``. ``tensor.shape[labeled_axis]`` must have the
        shape of ``len(labels)`` at runtime.

    Returns
    -------
    result: TensorVariable
        Result of addition of ``tensor`` and ``other_tensor``, along the ``labeled_axis`` dimension.  The ordering of
        the output will be ``labels + [label for label in other_labels if label not in labels]``. That is, ``labels``
        come first, followed by any new labels introduced by ``other_labels``.

    """
    labeled_axis = normalize_axis(tensor, labeled_axis)
    new_labels = [label for label in other_labels if label not in labels]
    combined_labels = type_cast(list[str], [*labels, *new_labels])

    # If there is no overlap at all, directly concatenate the two matrices -- there's no need to worry about the order
    # of things, or padding. This is equivalent to padding both out with zeros then adding them.
    if combined_labels == [*labels, *other_labels]:
        if isinstance(labeled_axis, int):
            return pt.concatenate([tensor, other_tensor], axis=labeled_axis)
        else:
            # In the case where we want to align multiple dimensions, use block_diag to accomplish padding on the last
            # two dimensions
            dims = [*[i for i in range(tensor.ndim) if i not in labeled_axis], *labeled_axis]
            return pt.linalg.block_diag(
                type_cast(TensorVariable, tensor.transpose(*dims)),
                type_cast(TensorVariable, other_tensor.transpose(*dims)),
            )
    # Otherwise, there are two possibilities. If all labels are the same, we might need to re-order one or both to get
    # them to agree. If *some* labels are the same, we will need to pad first, then potentially re-order. In any case,
    # the final step is just to add the padded and re-ordered tensors.
    fn = pad_and_reorder if isinstance(labeled_axis, int) else ndim_pad_and_reorder

    padded_tensor = fn(
        tensor,
        labels=type_cast(list[str], labels),
        ordered_labels=combined_labels,
        labeled_axis=labeled_axis,
    )
    padded_tensor.name = tensor.name

    padded_other_tensor = fn(
        other_tensor,
        labels=type_cast(list[str], other_labels),
        ordered_labels=combined_labels,
        labeled_axis=labeled_axis,
    )

    padded_other_tensor.name = other_tensor.name

    return padded_tensor + padded_other_tensor


def join_tensors_by_dim_labels(
    tensor: TensorVariable,
    other_tensor: TensorVariable,
    labels: list[str],
    other_labels: list[str],
    labeled_axis: int = -1,
    join_axis: int = -1,
    block_diag_join: bool = False,
) -> TensorVariable:
    labeled_axis = normalize_axis(tensor, labeled_axis)
    new_labels = [label for label in other_labels if label not in labels]
    combined_labels = [*labels, *new_labels]

    # Check for no overlap first. In this case, do a block_diagonal join, which implicitly results in padding zeros
    # everywhere they are needed -- no other sorting or padding necessary
    if combined_labels == [*labels, *other_labels]:
        res = pt.linalg.block_diag(tensor, other_tensor)
        new_shape = [
            shape_1 + shape_2 if (shape_1 is not None and shape_2 is not None) else None
            for shape_1, shape_2 in zip(tensor.type.shape, other_tensor.type.shape)
        ]
        return pt.specify_shape(res, new_shape)

    # Otherwise there is either total overlap or partial overlap. Let the padding and reordering function figure it out.
    tensor = ndim_pad_and_reorder(tensor, labels, combined_labels, labeled_axis)
    other_tensor = ndim_pad_and_reorder(other_tensor, other_labels, combined_labels, labeled_axis)

    if block_diag_join:
        new_shape = [
            shape_1 + shape_2 if (shape_1 is not None and shape_2 is not None) else None
            for shape_1, shape_2 in zip(tensor.type.shape, other_tensor.type.shape)
        ]
        res = pt.linalg.block_diag(tensor, other_tensor)
    else:
        new_shape = []
        join_axis_norm = normalize_axis(tensor, join_axis)
        for i, (shape_1, shape_2) in enumerate(zip(tensor.type.shape, other_tensor.type.shape)):
            if i == join_axis_norm:
                new_shape.append(
                    shape_1 + shape_2 if (shape_1 is not None and shape_2 is not None) else None
                )
            else:
                new_shape.append(shape_1 if shape_1 is not None else shape_2)
        res = pt.concatenate([tensor, other_tensor], axis=join_axis)

    return pt.specify_shape(res, new_shape)


def get_exog_dims_from_idata(exog_name, idata):
    if exog_name in idata.posterior.data_vars:
        exog_dims = idata.posterior[exog_name].dims[2:]
    elif exog_name in getattr(idata, "constant_data", []):
        exog_dims = idata.constant_data[exog_name].dims
    elif exog_name in getattr(idata, "mutable_data", []):
        exog_dims = idata.mutable_data[exog_name].dims
    else:
        exog_dims = None

    return exog_dims



"""Deserialize dictionaries into Python objects.

This is a two step process:

1. Determine if the data is of the correct type.
2. Deserialize the data into a python object.

Examples
--------
Make use of the already registered deserializers:

.. code-block:: python

    from pymc_extras.deserialize import deserialize

    prior_class_data = {
        "dist": "Normal",
        "kwargs": {"mu": 0, "sigma": 1}
    }
    prior = deserialize(prior_class_data)
    # Prior("Normal", mu=0, sigma=1)

Register custom class deserialization:

.. code-block:: python

    from pymc_extras.deserialize import register_deserialization

    class MyClass:
        def __init__(self, value: int):
            self.value = value

        def to_dict(self) -> dict:
            # Example of what the to_dict method might look like.
            return {"value": self.value}

    register_deserialization(
        is_type=lambda data: data.keys() == {"value"} and isinstance(data["value"], int),
        deserialize=lambda data: MyClass(value=data["value"]),
    )

Deserialize data into that custom class:

.. code-block:: python

    from pymc_extras.deserialize import deserialize

    data = {"value": 42}
    obj = deserialize(data)
    assert isinstance(obj, MyClass)


"""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

IsType = Callable[[Any], bool]
Deserialize = Callable[[Any], Any]


@dataclass
class Deserializer:
    """Object to store information required for deserialization.

    All deserializers should be stored via the :func:`register_deserialization` function
    instead of creating this object directly.

    Attributes
    ----------
    is_type : IsType
        Function to determine if the data is of the correct type.
    deserialize : Deserialize
        Function to deserialize the data.

    Examples
    --------
    .. code-block:: python

        from typing import Any

        class MyClass:
            def __init__(self, value: int):
                self.value = value

        from pymc_extras.deserialize import Deserializer

        def is_type(data: Any) -> bool:
            return data.keys() == {"value"} and isinstance(data["value"], int)

        def deserialize(data: dict) -> MyClass:
            return MyClass(value=data["value"])

        deserialize_logic = Deserializer(is_type=is_type, deserialize=deserialize)

    """

    is_type: IsType
    deserialize: Deserialize


DESERIALIZERS: list[Deserializer] = []


class DeserializableError(Exception):
    """Error raised when data cannot be deserialized."""

    def __init__(self, data: Any):
        self.data = data
        super().__init__(
            f"Couldn't deserialize {data}. Use register_deserialization to add a deserialization mapping."
        )


def deserialize(data: Any) -> Any:
    """Deserialize a dictionary into a Python object.

    Use the :func:`register_deserialization` function to add custom deserializations.

    Deserialization is a two step process due to the dynamic nature of the data:

    1. Determine if the data is of the correct type.
    2. Deserialize the data into a Python object.

    Each registered deserialization is checked in order until one is found that can
    deserialize the data. If no deserialization is found, a :class:`DeserializableError` is raised.

    A :class:`DeserializableError` is raised when the data fails to be deserialized
    by any of the registered deserializers.

    Parameters
    ----------
    data : Any
        The data to deserialize.

    Returns
    -------
    Any
        The deserialized object.

    Raises
    ------
    DeserializableError
        Raised when the data doesn't match any registered deserializations
        or fails to be deserialized.

    Examples
    --------
    Deserialize a :class:`pymc_extras.prior.Prior` object:

    .. code-block:: python

        from pymc_extras.deserialize import deserialize

        data = {"dist": "Normal", "kwargs": {"mu": 0, "sigma": 1}}
        prior = deserialize(data)
        # Prior("Normal", mu=0, sigma=1)

    """
    for mapping in DESERIALIZERS:
        try:
            is_type = mapping.is_type(data)
        except Exception:
            is_type = False

        if not is_type:
            continue

        try:
            return mapping.deserialize(data)
        except Exception as e:
            raise DeserializableError(data) from e
    else:
        raise DeserializableError(data)


def register_deserialization(is_type: IsType, deserialize: Deserialize) -> None:
    """Register an arbitrary deserialization.

    Use the :func:`deserialize` function to then deserialize data using all registered
    deserialize functions.

    Parameters
    ----------
    is_type : Callable[[Any], bool]
        Function to determine if the data is of the correct type.
    deserialize : Callable[[dict], Any]
        Function to deserialize the data of that type.

    Examples
    --------
    Register a custom class deserialization:

    .. code-block:: python

        from pymc_extras.deserialize import register_deserialization

        class MyClass:
            def __init__(self, value: int):
                self.value = value

            def to_dict(self) -> dict:
                # Example of what the to_dict method might look like.
                return {"value": self.value}

        register_deserialization(
            is_type=lambda data: data.keys() == {"value"} and isinstance(data["value"], int),
            deserialize=lambda data: MyClass(value=data["value"]),
        )

    Use that custom class deserialization:

    .. code-block:: python

        from pymc_extras.deserialize import deserialize

        data = {"value": 42}
        obj = deserialize(data)
        assert isinstance(obj, MyClass)

    """
    mapping = Deserializer(is_type=is_type, deserialize=deserialize)
    DESERIALIZERS.append(mapping)



from itertools import product

import numpy as np
import pandas as pd
import pymc as pm
import pytensor
import pytensor.tensor as pt
import pytest
import statsmodels.api as sm

from numpy.testing import assert_allclose
from pytensor.graph.basic import explicit_graph_inputs
from statsmodels.tsa.statespace.dynamic_factor import DynamicFactor

from pymc_extras.statespace.models.DFM import BayesianDynamicFactor
from pymc_extras.statespace.utils.constants import (
    ALL_STATE_AUX_DIM,
    ALL_STATE_DIM,
    AR_PARAM_DIM,
    ERROR_AR_PARAM_DIM,
    EXOG_STATE_DIM,
    FACTOR_DIM,
    LONG_MATRIX_NAMES,
    MATRIX_NAMES,
    OBS_STATE_AUX_DIM,
    OBS_STATE_DIM,
    SHORT_NAME_TO_LONG,
)
from tests.statespace.shared_fixtures import rng

floatX = pytensor.config.floatX


@pytest.fixture(scope="session")
def data():
    df = pd.read_csv(
        "tests/statespace/_data/statsmodels_macrodata_processed.csv",
        index_col=0,
        parse_dates=True,
    ).astype(floatX)
    df.index.freq = df.index.inferred_freq
    return df


def create_sm_test_values_mapping(
    test_values, data, k_factors, factor_order, error_order, error_var
):
    """Convert PyMC test values to statsmodels parameter format"""
    sm_test_values = {}

    # Factor loadings: PyMC shape (n_endog, k_factors) -> statsmodels individual params
    factor_loadings = test_values["factor_loadings"]
    all_pairs = product(data.columns, range(1, k_factors + 1))
    sm_test_values.update(
        {
            f"loading.f{factor_idx}.{endog_name}": value
            for (endog_name, factor_idx), value in zip(all_pairs, factor_loadings.ravel())
        }
    )

    # Factor AR coefficients: PyMC shape (k_factors, factor_order*k_factors) -> L{lag}.f{to}.f{from}
    if factor_order > 0 and "factor_ar" in test_values:
        factor_ar = test_values["factor_ar"]
        triplets = product(
            range(1, k_factors + 1), range(1, factor_order + 1), range(1, k_factors + 1)
        )
        sm_test_values.update(
            {
                f"L{lag}.f{to_factor}.f{from_factor}": factor_ar[
                    from_factor - 1, (lag - 1) * k_factors + (to_factor - 1)
                ]
                for from_factor, lag, to_factor in triplets
            }
        )

    # Error AR coefficients: PyMC shape (n_endog, error_order) -> L{lag}.e(var).e(var)
    if error_order > 0 and not error_var and "error_ar" in test_values:
        error_ar = test_values["error_ar"]
        pairs = product(enumerate(data.columns), range(1, error_order + 1))
        sm_test_values.update(
            {
                f"L{lag}.e({endog_name}).e({endog_name})": error_ar[endog_idx, lag - 1]
                for (endog_idx, endog_name), lag in pairs
            }
        )

    # Error AR coefficients: PyMC shape (n_endog, error_order * n_endog) -> L{lag}.e(var).e(var)
    elif error_order > 0 and error_var and "error_ar" in test_values:
        error_ar = test_values["error_ar"]
        triplets = product(
            enumerate(data.columns), range(1, error_order + 1), enumerate(data.columns)
        )
        sm_test_values.update(
            {
                f"L{lag}.e({from_endog_name}).e({to_endog_name})": error_ar[
                    from_endog_idx, (lag - 1) * data.shape[1] + to_endog_idx
                ]
                for (from_endog_idx, from_endog_name), lag, (
                    to_endog_idx,
                    to_endog_name,
                ) in triplets
            }
        )

    # Observation error variances:
    if "error_sigma" in test_values:
        error_sigma = test_values["error_sigma"]
        sm_test_values.update(
            {
                f"sigma2.{endog_name}": error_sigma[endog_idx]
                for endog_idx, endog_name in enumerate(data.columns)
            }
        )

    return sm_test_values


@pytest.mark.parametrize("k_factors", [1, 2])
@pytest.mark.parametrize("factor_order", [0, 1, 2])
@pytest.mark.parametrize("error_order", [0, 1, 2])
@pytest.mark.parametrize("error_var", [True, False])
@pytest.mark.filterwarnings("ignore::statsmodels.tools.sm_exceptions.EstimationWarning")
@pytest.mark.filterwarnings("ignore::FutureWarning")
def test_DFM_update_matches_statsmodels(data, k_factors, factor_order, error_order, error_var, rng):
    if error_var and (factor_order > 0 or error_order > 0):
        pytest.xfail(
            "Statsmodels may be doing something wrong with error_var=True and (factor_order > 0 or error_order > 0) [numpy.linalg.LinAlgError: 1-th leading minor of the array is not positive definite]"
        )

    mod = BayesianDynamicFactor(
        k_factors=k_factors,
        factor_order=factor_order,
        error_order=error_order,
        k_endog=data.shape[1],
        measurement_error=False,
        error_var=error_var,
        verbose=False,
    )
    sm_dfm = DynamicFactor(
        endog=data,
        k_factors=k_factors,
        factor_order=factor_order,
        error_order=error_order,
        error_var=error_var,
    )

    # Generate test values for PyMC model
    test_values = {}
    test_values["x0"] = rng.normal(size=mod.k_states)
    test_values["P0"] = np.eye(mod.k_states)
    test_values["factor_loadings"] = rng.normal(size=(data.shape[1], k_factors))

    if factor_order > 0:
        test_values["factor_ar"] = rng.normal(size=(k_factors, factor_order * k_factors))

    if error_order > 0 and error_var:
        test_values["error_ar"] = rng.normal(size=(data.shape[1], error_order * data.shape[1]))
    elif error_order > 0 and not error_var:
        test_values["error_ar"] = rng.normal(size=(data.shape[1], error_order))

    test_values["error_sigma"] = rng.beta(1, 1, size=data.shape[1])

    # Convert to statsmodels format
    sm_test_values = create_sm_test_values_mapping(
        test_values, data, k_factors, factor_order, error_order, error_var
    )

    x0 = test_values["x0"]
    P0 = test_values["P0"]

    sm_dfm.initialize_known(initial_state=x0, initial_state_cov=P0)
    sm_dfm.fit_constrained({name: sm_test_values[name] for name in sm_dfm.param_names})

    # Get PyMC matrices
    matrices = mod._unpack_statespace_with_placeholders()
    inputs = list(explicit_graph_inputs(matrices))
    input_names = [x.name for x in inputs]

    f_matrices = pytensor.function(inputs, matrices)
    test_values_subset = {name: test_values[name] for name in input_names if name in test_values}

    pymc_matrices = f_matrices(**test_values_subset)

    sm_matrices = [sm_dfm.ssm[name] for name in LONG_MATRIX_NAMES[2:]]

    # Compare matrices (skip x0 and P0)
    for matrix, sm_matrix, name in zip(pymc_matrices[2:], sm_matrices, LONG_MATRIX_NAMES[2:]):
        assert_allclose(matrix, sm_matrix, err_msg=f"{name} does not match")


def unpack_statespace(ssm):
    return [ssm[SHORT_NAME_TO_LONG[x]] for x in MATRIX_NAMES]


def unpack_symbolic_matrices_with_params(mod, param_dict, data_dict=None, mode="FAST_COMPILE"):
    inputs = list(mod._name_to_variable.values())
    if data_dict is not None:
        inputs += list(mod._name_to_data.values())
    else:
        data_dict = {}

    f_matrices = pytensor.function(
        inputs,
        unpack_statespace(mod.ssm),
        on_unused_input="raise",
        mode=mode,
    )

    return f_matrices(**param_dict, **data_dict)


def simulate_from_numpy_model(
    mod, rng, param_dict, data_dict=None, steps=100, state_shocks=None, measurement_shocks=None
):
    x0, P0, c, d, T, Z, R, H, Q = unpack_symbolic_matrices_with_params(mod, param_dict, data_dict)
    k_endog = mod.k_endog
    k_states = mod.k_states
    k_posdef = mod.k_posdef

    x = np.zeros((steps, k_states))
    y = np.zeros((steps, k_endog))

    x[0] = x0
    y[0] = (Z @ x0).squeeze() if Z.ndim == 2 else (Z[0] @ x0).squeeze()

    if not np.allclose(H, 0):
        y[0] += rng.multivariate_normal(mean=np.zeros(1), cov=H).squeeze()

    for t in range(1, steps):
        if k_posdef > 0:
            innov = R @ rng.multivariate_normal(mean=np.zeros(k_posdef), cov=Q)
        else:
            innov = 0

        if not np.allclose(H, 0):
            error = measurement_shocks[t - 1]
        else:
            error = 0

        x[t] = c + T @ x[t - 1] + innov
        if Z.ndim == 2:
            y[t] = (d + Z @ x[t] + error).squeeze()
        else:
            y[t] = (d + Z[t] @ x[t] + error).squeeze()

    return x, y.squeeze()


@pytest.mark.parametrize("n_obs,n_runs", [(100, 200)])
def test_DFM_exog_betas_random_walk(n_obs, n_runs):
    rng = np.random.default_rng(123)
    dfm_mod = BayesianDynamicFactor(
        k_factors=1,
        factor_order=1,
        k_endog=2,
        error_order=1,
        error_var=False,
        k_exog=2,
        shared_exog_states=False,
        exog_innovations=True,
        error_cov_type="diagonal",
        measurement_error=False,
    )

    # Arbitrary Parameters
    param_dict = {
        "factor_loadings": np.array([[0.9], [0.8]]),
        "factor_ar": np.array([[0.5]]),
        "error_ar": np.array([[0.4], [0.3]]),
        "error_sigma": np.array([0.1, 0.2]),
        "P0": np.eye(dfm_mod.k_states),
        "x0": np.zeros(dfm_mod.k_states - dfm_mod.k_exog * dfm_mod.k_endog),
        "beta": np.array([0.3, 0.5, 1, 2]),
        "beta_sigma": np.array([1, 2, 3, 4]) ** 0.5,
    }
    data_dict = {"exog_data": np.random.normal(size=(n_obs, 2))}

    # Run multiple sims
    betas_t1, betas_t100 = [], []
    k_exog_states = dfm_mod.k_exog * dfm_mod.k_endog

    for _ in range(n_runs):
        x_traj, _ = simulate_from_numpy_model(dfm_mod, rng, param_dict, data_dict, steps=n_obs)
        beta_traj = x_traj[:, -k_exog_states:]
        betas_t1.append(beta_traj[1, :])
        betas_t100.append(beta_traj[-1, :])

    betas_t1 = np.array(betas_t1)
    betas_t100 = np.array(betas_t100)

    var_t1 = betas_t1.var(axis=0)
    var_t100 = betas_t100.var(axis=0)

    assert np.all(
        var_t100 > var_t1
    ), f"Expected variance at T=100 > T=1, got {var_t1} vs {var_t100}"


@pytest.mark.parametrize("shared", [True, False])
def test_DFM_exog_shared_vs_not(shared):
    rng = np.random.default_rng(123)

    n_obs = 50
    k_exog = 2
    k_endog = 2

    # Dummy exogenous data
    exog = rng.normal(size=(n_obs, k_exog))

    dfm_mod = BayesianDynamicFactor(
        k_factors=1,
        factor_order=1,
        k_endog=k_endog,
        error_order=1,
        k_exog=k_exog,
        shared_exog_states=shared,
        exog_innovations=False,
        error_cov_type="diagonal",
        measurement_error=False,
    )

    k_exog_states = dfm_mod.k_exog * dfm_mod.k_endog if not shared else dfm_mod.k_exog

    if shared:
        beta = np.array([0.3, 0.5])
    else:
        beta = np.array([0.3, 0.5, 1.0, 2.0])

    param_dict = {
        "factor_loadings": np.array([[0.9], [0.8]]),
        "factor_ar": np.array([[0.5]]),
        "error_ar": np.array([[0.4], [0.3]]),
        "error_sigma": np.array([0.1, 0.2]),
        "P0": np.eye(dfm_mod.k_states),
        "x0": np.zeros(dfm_mod.k_states - k_exog_states),
        "beta": beta,
    }

    data_dict = {"exog_data": exog}

    # Simulate trajectory
    x_traj, y_traj = simulate_from_numpy_model(dfm_mod, rng, param_dict, data_dict, steps=n_obs)

    # Test 1: Check hidden states
    # Extract exogenous hidden states at time t=10
    t = 10
    exog_states_start = dfm_mod.k_states - k_exog_states
    exog_states_end = dfm_mod.k_states
    exog_hidden_states = x_traj[t, exog_states_start:exog_states_end]

    if shared:
        # When shared=True, there should be k_exog states total
        assert len(exog_hidden_states) == k_exog
    else:
        # When shared=False, there should be k_exog * k_endog states
        assert len(exog_hidden_states) == k_exog * k_endog
        # Each endogenous variable has its own set of exogenous states
        exog_states_reshaped = exog_hidden_states.reshape(k_endog, k_exog)
        assert not np.allclose(exog_states_reshaped[0], exog_states_reshaped[1])

    # Test 2: Check observed contributions
    exog_t = exog[t]

    if shared:
        # All endogenous variables get the same beta * data contribution
        contributions = [beta @ exog_t for _ in range(k_endog)]
        assert np.allclose(
            contributions[0], contributions[1]
        ), "Expected same contribution for all endog when shared=True"
    else:
        # Each endogenous variable gets different beta * data
        beta_reshaped = beta.reshape(k_endog, k_exog)
        contributions = [beta_reshaped[i] @ exog_t for i in range(k_endog)]
        # Check that contributions are different
        assert not np.allclose(
            contributions[0], contributions[1]
        ), f"Expected different contributions, got {contributions}"


class TestDFMConfiguration:
    def test_static_factor_no_ar_no_exog_diagonal_error(self):
        mod = BayesianDynamicFactor(
            k_factors=1,
            factor_order=0,
            k_endog=3,
            endog_names=["y0", "y1", "y2"],
            error_order=0,
            error_var=False,
            error_cov_type="diagonal",
            measurement_error=False,
            verbose=False,
        )

        expected_param_names = ["x0", "P0", "factor_loadings", "error_sigma"]
        expected_param_dims = {
            "x0": (ALL_STATE_DIM,),
            "P0": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
            "factor_loadings": (OBS_STATE_DIM, FACTOR_DIM),
            "error_sigma": (OBS_STATE_DIM,),
        }
        expected_coords = {
            OBS_STATE_DIM: ["y0", "y1", "y2"],
            ALL_STATE_DIM: ["L0.factor_0"],
            ALL_STATE_AUX_DIM: ["L0.factor_0"],
            FACTOR_DIM: ["factor_1"],
        }

        assert mod.param_names == expected_param_names
        assert mod.param_dims == expected_param_dims
        for k, v in expected_coords.items():
            assert mod.coords[k] == v
        assert mod.state_names == ["L0.factor_0"]
        assert mod.observed_states == ["y0", "y1", "y2"]
        assert mod.shock_names == ["factor_shock_0"]

    def test_dynamic_factor_ar1_error_diagonal_error(self):
        k_factors = 2
        factor_order = 2
        k_endog = 3
        error_order = 1
        error_var = False

        mod = BayesianDynamicFactor(
            k_factors=k_factors,
            factor_order=factor_order,
            k_endog=k_endog,
            endog_names=["y0", "y1", "y2"],
            error_order=error_order,
            error_var=error_var,
            error_cov_type="diagonal",
            measurement_error=True,
            verbose=False,
        )
        expected_param_names = [
            "x0",
            "P0",
            "factor_loadings",
            "factor_ar",
            "error_ar",
            "error_sigma",
            "sigma_obs",
        ]
        expected_param_dims = {
            "x0": (ALL_STATE_DIM,),
            "P0": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
            "factor_loadings": (OBS_STATE_DIM, FACTOR_DIM),
            "factor_ar": (FACTOR_DIM, AR_PARAM_DIM),
            "error_ar": (OBS_STATE_DIM, ERROR_AR_PARAM_DIM),
            "error_sigma": (OBS_STATE_DIM,),
            "sigma_obs": (OBS_STATE_DIM,),
        }
        expected_coords = {
            OBS_STATE_DIM: ["y0", "y1", "y2"],
            ALL_STATE_DIM: [
                "L0.factor_0",
                "L1.factor_0",
                "L0.factor_1",
                "L1.factor_1",
                "L0.error_0",
                "L0.error_1",
                "L0.error_2",
            ],
            ALL_STATE_AUX_DIM: [
                "L0.factor_0",
                "L1.factor_0",
                "L0.factor_1",
                "L1.factor_1",
                "L0.error_0",
                "L0.error_1",
                "L0.error_2",
            ],
            FACTOR_DIM: ["factor_1", "factor_2"],
            AR_PARAM_DIM: list(range(1, k_factors * max(factor_order, 1) + 1)),
            ERROR_AR_PARAM_DIM: list(range(1, (error_order * k_endog) + 1))
            if error_var
            else list(range(1, error_order + 1)),
        }

        assert mod.param_names == expected_param_names
        assert mod.param_dims == expected_param_dims
        for k, v in expected_coords.items():
            assert mod.coords[k] == v
        assert len(mod.state_names) == k_factors * max(factor_order, 1) + k_endog * error_order
        assert mod.observed_states == ["y0", "y1", "y2"]
        assert len(mod.shock_names) == k_factors + k_endog

    def test_dynamic_factor_ar2_error_var_unstructured(self):
        k_factors = 1
        factor_order = 1
        k_endog = 3
        error_order = 2
        error_var = True
        mod = BayesianDynamicFactor(
            k_factors=k_factors,
            factor_order=factor_order,
            k_endog=k_endog,
            endog_names=["y0", "y1", "y2"],
            error_order=error_order,
            error_var=error_var,
            error_cov_type="unstructured",
            measurement_error=True,
            verbose=False,
        )
        expected_param_names = [
            "x0",
            "P0",
            "factor_loadings",
            "factor_ar",
            "error_ar",
            "error_cov",
            "sigma_obs",
        ]
        expected_param_dims = {
            "x0": (ALL_STATE_DIM,),
            "P0": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
            "factor_loadings": (OBS_STATE_DIM, FACTOR_DIM),
            "factor_ar": (FACTOR_DIM, AR_PARAM_DIM),
            "error_ar": (OBS_STATE_DIM, ERROR_AR_PARAM_DIM),
            "error_cov": (OBS_STATE_DIM, OBS_STATE_AUX_DIM),
            "sigma_obs": (OBS_STATE_DIM,),
        }
        expected_coords = {
            OBS_STATE_DIM: ["y0", "y1", "y2"],
            ALL_STATE_DIM: [
                "L0.factor_0",
                "L0.error_0",
                "L1.error_0",
                "L0.error_1",
                "L1.error_1",
                "L0.error_2",
                "L1.error_2",
            ],
            ALL_STATE_AUX_DIM: [
                "L0.factor_0",
                "L0.error_0",
                "L1.error_0",
                "L0.error_1",
                "L1.error_1",
                "L0.error_2",
                "L1.error_2",
            ],
            FACTOR_DIM: ["factor_1"],
            AR_PARAM_DIM: list(range(1, k_factors * max(factor_order, 1) + 1)),
            ERROR_AR_PARAM_DIM: list(range(1, (error_order * k_endog) + 1))
            if error_var
            else list(range(1, error_order + 1)),
        }

        assert mod.param_names == expected_param_names
        assert mod.param_dims == expected_param_dims
        for k, v in expected_coords.items():
            assert mod.coords[k] == v
        assert len(mod.state_names) == k_factors * max(factor_order, 1) + k_endog * error_order
        assert mod.observed_states == ["y0", "y1", "y2"]
        assert len(mod.shock_names) == k_factors + k_endog

    def test_exog_shared_exog_states_exog_innovations(self):
        k_factors = 2
        factor_order = 1
        k_endog = 3
        error_order = 1
        k_exog = 2
        error_var = False
        shared_exog_states = True
        mod = BayesianDynamicFactor(
            k_factors=k_factors,
            factor_order=factor_order,
            k_endog=k_endog,
            endog_names=["y0", "y1", "y2"],
            error_order=error_order,
            error_var=error_var,
            k_exog=2,
            exog_names=["x0", "x1"],
            shared_exog_states=shared_exog_states,
            exog_innovations=True,
            error_cov_type="diagonal",
            measurement_error=True,
            verbose=False,
        )
        expected_param_names = [
            "x0",
            "P0",
            "factor_loadings",
            "factor_ar",
            "error_ar",
            "error_sigma",
            "sigma_obs",
            "beta",
            "beta_sigma",
        ]
        expected_param_dims = {
            "x0": (ALL_STATE_DIM,),
            "P0": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
            "factor_loadings": (OBS_STATE_DIM, FACTOR_DIM),
            "factor_ar": (FACTOR_DIM, AR_PARAM_DIM),
            "error_ar": (OBS_STATE_DIM, ERROR_AR_PARAM_DIM),
            "error_sigma": (OBS_STATE_DIM,),
            "sigma_obs": (OBS_STATE_DIM,),
            "beta": (EXOG_STATE_DIM,),
            "beta_sigma": (EXOG_STATE_DIM,),
        }
        expected_coords = {
            OBS_STATE_DIM: ["y0", "y1", "y2"],
            ALL_STATE_DIM: [
                "L0.factor_0",
                "L0.factor_1",
                "L0.error_0",
                "L0.error_1",
                "L0.error_2",
                "beta_x0[shared]",
                "beta_x1[shared]",
            ],
            ALL_STATE_AUX_DIM: [
                "L0.factor_0",
                "L0.factor_1",
                "L0.error_0",
                "L0.error_1",
                "L0.error_2",
                "beta_x0[shared]",
                "beta_x1[shared]",
            ],
            FACTOR_DIM: ["factor_1", "factor_2"],
            AR_PARAM_DIM: list(range(1, k_factors * max(factor_order, 1) + 1)),
            ERROR_AR_PARAM_DIM: list(range(1, (error_order * k_endog) + 1))
            if error_var
            else list(range(1, error_order + 1)),
            EXOG_STATE_DIM: list(range(1, k_exog + 1))
            if shared_exog_states
            else list(range(1, k_exog * k_endog + 1)),
        }

        assert mod.param_names == expected_param_names
        assert mod.param_dims == expected_param_dims
        for k, v in expected_coords.items():
            assert mod.coords[k] == v
        assert len(mod.state_names) == k_factors * max(factor_order, 1) + k_endog * error_order + (
            k_exog if shared_exog_states else k_exog * k_endog
        )
        assert mod.observed_states == ["y0", "y1", "y2"]
        assert len(mod.shock_names) == k_factors + k_endog + (
            k_exog if shared_exog_states else k_exog * k_endog
        )

    def test_exog_not_shared_no_exog_innovations(self):
        k_factors = 1
        factor_order = 2
        k_endog = 3
        error_order = 1
        k_exog = 1
        error_var = False
        shared_exog_states = False
        mod = BayesianDynamicFactor(
            k_factors=k_factors,
            factor_order=factor_order,
            k_endog=k_endog,
            endog_names=["y0", "y1", "y2"],
            error_order=error_order,
            error_var=error_var,
            k_exog=k_exog,
            exog_names=["x0"],
            shared_exog_states=shared_exog_states,
            exog_innovations=False,
            error_cov_type="scalar",
            measurement_error=False,
            verbose=False,
        )
        expected_param_names = [
            "x0",
            "P0",
            "factor_loadings",
            "factor_ar",
            "error_ar",
            "error_sigma",
            "beta",
        ]
        expected_param_dims = {
            "x0": (ALL_STATE_DIM,),
            "P0": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
            "factor_loadings": (OBS_STATE_DIM, FACTOR_DIM),
            "factor_ar": (FACTOR_DIM, AR_PARAM_DIM),
            "error_ar": (OBS_STATE_DIM, ERROR_AR_PARAM_DIM),
            "error_sigma": (),
            "beta": (EXOG_STATE_DIM,),
        }
        expected_coords = {
            OBS_STATE_DIM: ["y0", "y1", "y2"],
            ALL_STATE_DIM: [
                "L0.factor_0",
                "L1.factor_0",
                "L0.error_0",
                "L0.error_1",
                "L0.error_2",
                "beta_x0[y0]",
                "beta_x0[y1]",
                "beta_x0[y2]",
            ],
            ALL_STATE_AUX_DIM: [
                "L0.factor_0",
                "L1.factor_0",
                "L0.error_0",
                "L0.error_1",
                "L0.error_2",
                "beta_x0[y0]",
                "beta_x0[y1]",
                "beta_x0[y2]",
            ],
            FACTOR_DIM: ["factor_1"],
            AR_PARAM_DIM: list(range(1, k_factors * max(factor_order, 1) + 1)),
            ERROR_AR_PARAM_DIM: list(range(1, (error_order * k_endog) + 1))
            if error_var
            else list(range(1, error_order + 1)),
            EXOG_STATE_DIM: list(range(1, k_exog + 1))
            if shared_exog_states
            else list(range(1, k_exog * k_endog + 1)),
        }

        assert mod.param_names == expected_param_names
        assert mod.param_dims == expected_param_dims
        for k, v in expected_coords.items():
            assert mod.coords[k] == v
        assert len(mod.state_names) == k_factors * max(factor_order, 1) + k_endog * error_order + (
            k_exog if shared_exog_states else k_exog * k_endog
        )
        assert mod.observed_states == ["y0", "y1", "y2"]
        assert len(mod.shock_names) == k_factors + k_endog + (
            k_exog if shared_exog_states else k_exog * k_endog
        )



import numpy as np
import pytensor

from numpy.testing import assert_allclose
from pytensor import config

from pymc_extras.statespace.models import structural as st
from tests.statespace.models.structural.conftest import _assert_basic_coords_correct
from tests.statespace.test_utilities import simulate_from_numpy_model

ATOL = 1e-8 if config.floatX.endswith("64") else 1e-4
RTOL = 0 if config.floatX.endswith("64") else 1e-6


def test_level_trend_model(rng):
    mod = st.LevelTrendComponent(order=2, innovations_order=0)
    params = {"initial_level_trend": [0.0, 1.0]}
    x, y = simulate_from_numpy_model(mod, rng, params)

    assert_allclose(np.diff(y), 1, atol=ATOL, rtol=RTOL)

    # Check coords
    mod = mod.build(verbose=False)
    _assert_basic_coords_correct(mod)
    assert mod.coords["state_level_trend"] == ["level", "trend"]


def test_level_trend_multiple_observed_construction():
    mod = st.LevelTrendComponent(
        order=2, innovations_order=1, observed_state_names=["data_1", "data_2", "data_3"]
    )
    mod = mod.build(verbose=False)
    assert mod.k_endog == 3
    assert mod.k_states == 6
    assert mod.k_posdef == 3

    assert mod.coords["state_level_trend"] == ["level", "trend"]
    assert mod.coords["endog_level_trend"] == ["data_1", "data_2", "data_3"]

    assert mod.state_names == [
        "level[data_1]",
        "trend[data_1]",
        "level[data_2]",
        "trend[data_2]",
        "level[data_3]",
        "trend[data_3]",
    ]
    assert mod.shock_names == ["level[data_1]", "level[data_2]", "level[data_3]"]

    Z, T, R = pytensor.function(
        [], [mod.ssm["design"], mod.ssm["transition"], mod.ssm["selection"]], mode="FAST_COMPILE"
    )()

    np.testing.assert_allclose(
        Z,
        np.array(
            [
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            ]
        ),
    )

    np.testing.assert_allclose(
        T,
        np.array(
            [
                [1.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            ]
        ),
    )

    np.testing.assert_allclose(
        R,
        np.array(
            [
                [1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0],
            ]
        ),
    )


def test_level_trend_multiple_shared_construction():
    mod = st.LevelTrendComponent(
        order=2, innovations_order=1, observed_state_names=["data_1", "data_2"], share_states=True
    )
    mod = mod.build(verbose=False)

    assert mod.k_endog == 2
    assert mod.k_states == 2
    assert mod.k_posdef == 1

    assert mod.coords["state_level_trend"] == ["level", "trend"]
    assert mod.coords["endog_level_trend"] == ["data_1", "data_2"]

    assert mod.state_names == [
        "level[level_trend_shared]",
        "trend[level_trend_shared]",
    ]
    assert mod.shock_names == ["level[level_trend_shared]"]

    Z, T, R = pytensor.function(
        [], [mod.ssm["design"], mod.ssm["transition"], mod.ssm["selection"]], mode="FAST_COMPILE"
    )()

    np.testing.assert_allclose(
        Z,
        np.array(
            [
                [1.0, 0.0],
                [1.0, 0.0],
            ]
        ),
    )

    np.testing.assert_allclose(T, np.array([[1.0, 1.0], [0.0, 1.0]]))

    np.testing.assert_allclose(R, np.array([[1.0], [0.0]]))


def test_level_trend_multiple_observed(rng):
    mod = st.LevelTrendComponent(
        order=2, innovations_order=0, observed_state_names=["data_1", "data_2", "data_3"]
    )
    params = {"initial_level_trend": np.array([[0.0, 1.0], [0.0, 2.0], [0.0, 3.0]])}

    x, y = simulate_from_numpy_model(mod, rng, params)
    assert (np.diff(y, axis=0) == np.array([[1.0, 2.0, 3.0]])).all().all()
    assert (np.diff(x, axis=0) == np.array([[1.0, 0.0, 2.0, 0.0, 3.0, 0.0]])).all().all()


def test_level_trend_multiple_shared_observed(rng):
    mod = st.LevelTrendComponent(
        order=2,
        innovations_order=0,
        observed_state_names=["data_1", "data_2", "data_3"],
        share_states=True,
    )
    params = {"initial_level_trend": np.array([10.0, 0.1])}
    x, y = simulate_from_numpy_model(mod, rng, params)
    np.testing.assert_allclose(y[:, 0], y[:, 1])
    np.testing.assert_allclose(y[:, 0], y[:, 2])


def test_add_level_trend_with_different_observed():
    mod_1 = st.LevelTrendComponent(
        name="ll", order=2, innovations_order=[0, 1], observed_state_names=["data_1"]
    )
    mod_2 = st.LevelTrendComponent(
        name="grw", order=1, innovations_order=[1], observed_state_names=["data_2"]
    )

    mod = (mod_1 + mod_2).build(verbose=False)
    assert mod.k_endog == 2
    assert mod.k_states == 3
    assert mod.k_posdef == 2

    assert mod.coords["state_ll"] == ["level", "trend"]
    assert mod.coords["state_grw"] == ["level"]

    assert mod.state_names == ["level[data_1]", "trend[data_1]", "level[data_2]"]
    assert mod.shock_names == ["trend[data_1]", "level[data_2]"]

    Z, T, R = pytensor.function(
        [], [mod.ssm["design"], mod.ssm["transition"], mod.ssm["selection"]], mode="FAST_COMPILE"
    )()

    np.testing.assert_allclose(
        Z,
        np.array(
            [
                [1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0],
            ]
        ),
    )

    np.testing.assert_allclose(
        T,
        np.array(
            [
                [1.0, 1.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
            ]
        ),
    )

    np.testing.assert_allclose(
        R,
        np.array(
            [
                [0.0, 0.0],
                [1.0, 0.0],
                [0.0, 1.0],
            ]
        ),
    )


def test_mixed_shared_and_not_shared():
    mod_1 = st.LevelTrendComponent(
        name="individual",
        order=2,
        innovations_order=[0, 1],
        observed_state_names=["data_1", "data_2"],
    )
    mod_2 = st.LevelTrendComponent(
        name="joint",
        order=2,
        innovations_order=[1, 1],
        observed_state_names=["data_1", "data_2"],
        share_states=True,
    )

    mod = (mod_1 + mod_2).build(verbose=False)

    assert mod.k_endog == 2
    assert mod.k_states == 6
    assert mod.k_posdef == 4

    assert mod.state_names == [
        "level[data_1]",
        "trend[data_1]",
        "level[data_2]",
        "trend[data_2]",
        "level[joint_shared]",
        "trend[joint_shared]",
    ]

    assert mod.shock_names == [
        "trend[data_1]",
        "trend[data_2]",
        "level[joint_shared]",
        "trend[joint_shared]",
    ]

    Z, T, R = pytensor.function(
        [], [mod.ssm["design"], mod.ssm["transition"], mod.ssm["selection"]], mode="FAST_COMPILE"
    )()

    np.testing.assert_allclose(
        Z, np.array([[1.0, 0.0, 0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 1.0, 0.0, 1.0, 0.0]])
    )

    np.testing.assert_allclose(
        T,
        np.array(
            [
                [1.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            ]
        ),
    )

    np.testing.assert_allclose(
        R,
        np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        ),
    )
